from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.db.models.constraints import UniqueConstraint
from localized_fields.fields import LocalizedField

from caluma.caluma_core.models import SlugModel, UUIDModel

from .simple_table import BaseStartingObject


class AnalyticsTable(SlugModel):
    STARTING_OBJECT_CHOICES = BaseStartingObject.as_choices()

    meta = models.JSONField(default=dict)

    disable_visibilities = models.BooleanField(default=False)
    name = LocalizedField(blank=False, null=False, required=False)
    starting_object = models.CharField(max_length=250, choices=STARTING_OBJECT_CHOICES)

    def get_starting_object(self, info):
        return BaseStartingObject.get_object(self.starting_object, info)


class AnalyticsField(UUIDModel):
    alias = models.CharField(max_length=100)
    meta = models.JSONField(default=dict)

    data_source = models.TextField()
    table = models.ForeignKey(
        AnalyticsTable, on_delete=models.CASCADE, related_name="fields"
    )
    filters = ArrayField(models.TextField(), blank=True, null=True, default=list)

    class Meta:
        constraints = [
            UniqueConstraint(
                name="unique_data_source", fields=["table", "data_source"]
            ),
            UniqueConstraint(name="unique_alias", fields=["table", "alias"]),
        ]



# Generated by Django 2.2.8 on 2020-02-10 09:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0016_auto_20200113_0727")]

    operations = [
        migrations.AlterField(
            model_name="case",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="case",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="case",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="flow",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="flow",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="flow",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalcase",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalcase",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalcase",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalflow",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalflow",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalflow",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicaltask",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicaltask",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicaltask",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicaltaskflow",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicaltaskflow",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicaltaskflow",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalworkflow",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalworkflow",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalworkflow",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalworkitem",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalworkitem",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalworkitem",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="task",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="task",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="taskflow",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="taskflow",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="taskflow",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="workitem",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="workitem",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="workitem",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
    ]



# Generated by Django 2.2.8 on 2020-01-13 07:27

import django.contrib.postgres.indexes
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0015_add_work_item_skipped")]

    operations = [
        migrations.RemoveIndex(model_name="case", name="workflow_ca_meta_08775a_gin"),
        migrations.RemoveIndex(model_name="task", name="workflow_ta_meta_6864a4_gin"),
        migrations.RemoveIndex(
            model_name="workflow", name="workflow_wo_meta_2ac517_gin"
        ),
        migrations.RemoveIndex(
            model_name="workitem", name="workflow_wo_address_679262_gin"
        ),
        migrations.RemoveIndex(
            model_name="workitem", name="workflow_wo_assigne_76d859_gin"
        ),
        migrations.RemoveIndex(
            model_name="workitem", name="workflow_wo_meta_2704a2_gin"
        ),
        migrations.AddIndex(
            model_name="case",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="caluma_work_meta_5cd3f3_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="task",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="caluma_work_meta_25e8ed_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="workflow",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="caluma_work_meta_42c1ce_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="workitem",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["addressed_groups"], name="caluma_work_address_23d0a8_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="workitem",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["assigned_users"], name="caluma_work_assigne_7880f0_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="workitem",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="caluma_work_meta_f6fc45_gin"
            ),
        ),
    ]



from factory import Faker, django

from . import models


class AccessLogFactory(django.DjangoModelFactory):
    timestamp = None
    username = Faker("user_name")
    query = ""
    operation = "query"
    operation_name = None
    selection = None
    variables = None
    status_code = 200
    has_error = False

    class Meta:
        model = models.AccessLog



from django_filters.rest_framework import FilterSet

from ..caluma_core.filters import MetaFilterSet, SearchFilter, SlugMultipleChoiceFilter
from ..caluma_core.ordering import AttributeOrderingFactory, MetaFieldOrdering
from . import models


class AnalyticsTableFilterSet(MetaFilterSet):
    search = SearchFilter(fields=("slug", "name"))
    slugs = SlugMultipleChoiceFilter(field_name="slug")

    class Meta:
        model = models.AnalyticsTable
        fields = (
            "slug",
            "name",
        )


class AnalyticsFieldFilterSet(MetaFilterSet):
    search = SearchFilter(fields=("alias", "data_source"))
    slugs = SlugMultipleChoiceFilter(field_name="slug")

    class Meta:
        model = models.AnalyticsField
        fields = (
            "alias",
            "table",
        )


class AnalyticsTableOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.AnalyticsTable,
        fields=[
            "created_at",
            "modified_at",
            "created_by_user",
            "created_by_group",
            "modified_by_user",
            "modified_by_group",
            "slug",
            "name",
        ],
    )

    class Meta:
        model = models.AnalyticsTable
        fields = ("meta", "attribute")



import json

from django.conf import settings
from django.http import Http404, JsonResponse
from watchman import settings as watchman_settings, views as watchman_views

from caluma.caluma_user.models import AnonymousUser


def _remove_keys(dictionary, keys):
    """Recursively remove given keys from dictionary."""
    for key in keys:  # delete keys
        if key in dictionary:
            del dictionary[key]
    for value in dictionary.values():  # retrieve nested values
        if isinstance(value, dict):
            _remove_keys(value, keys)  # apply to dict
        if isinstance(value, list):
            [_remove_keys(v, keys) for v in value]  # apply to list


def health_check_status(request):
    """Modify django-watchman view to remove error / stacktrace output.

    Available django-watchman views:
    watchman.views.status
    watchman.views.bare_status
    """
    # if health endpoint not enabled, return 404 response
    if not settings.ENABLE_HEALTHZ_ENDPOINT:
        raise Http404("Healthz endpoint not enabled.")

    # set anonymous user for requests to health endpoint
    # caluma expects user to be set on requests
    request.user = AnonymousUser()

    # Get view response from django-watchman
    response = watchman_views.status(request)
    json_data = json.loads(response.content.decode("utf-8"))

    # Remove unwanted keys e.g. 'error', 'stacktrace'
    if response.status_code == watchman_settings.WATCHMAN_ERROR_CODE:
        _remove_keys(json_data, ["error", "stacktrace"])
        return JsonResponse(json_data, status=response.status_code)

    return response



from __future__ import annotations  # self-referencing annotations

import hashlib
import textwrap
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple, Union
from uuid import uuid4

from django.db import connection
from django.utils.text import slugify


def _make_name(value, name_hint=None):
    prefix = name_hint if name_hint else "p"
    length = 5 if name_hint else 12

    suffix = hashlib.md5(str(value).encode("utf-8")).hexdigest()
    return f"{prefix}_{suffix[:length]}"


@dataclass
class Query:
    """Represent an SQL query (might be a subquery).

    Note: The `from` specification might in turn again
    reference another Query object, thus building a subquery
    structure.
    """

    from_: Union[str, Query]
    distinct_on: Optional[str] = field(default=None)
    select: List[Tuple(str, str)] = field(default_factory=list)  # expression -> alias
    outer_ref: Optional[Tuple(str, str)] = field(
        default=None
    )  # Pair of outer identifier, inner identifier. Used for JOINing

    joins: List[Tuple(Union[str, Query], str)] = field(
        default_factory=list
    )  # table/subquery, join expr

    filters: List[str] = field(default_factory=list)
    order_by: List[str] = field(default_factory=list)
    limit: Optional[int] = field(default=None)
    params: Dict[str, str] = field(default_factory=dict)
    with_queries: Dict[str, str] = field(
        default_factory=dict
    )  # with_queries are CTEs, used for referencing visibility-filtered tables

    is_plain_cte: bool = field(default=False)

    def makeparam(self, value, name_hint=None):
        """Return a parameter name for the given value.

        The parameter's value is automatically registered and
        can then be used in the generated SQL.

        If you pass a name_hint, it is used as a prefix
        to better identify the params when debugging the query
        """
        identifier = _make_name(value, name_hint)
        self.params[identifier] = value
        return f"%({identifier})s"

    def outer_alias(self):
        """Alias of the table / subquery, as it is referred to from *outside*.

        Used by outer queries to know how to alias our result expression.
        """

        if isinstance(self.from_, str):
            return self.from_

        subname_data = f"""
            {self.order_by}
            {self.filters}
        """

        # return f"sub_{self.from_.outer_alias()}_{ob}"
        return _make_name(subname_data, self.from_.outer_alias())

    def self_alias(self):
        """Alias of the table / subquery, as it is referred to from within.

        This shall be used for fully-qualified references to fields within
        this query.
        """
        if isinstance(self.from_, str):
            return self.from_
        return self.from_.outer_alias()

    def add_field_filter(self, alias, filter_values):
        """Apply the given filters to this field.

        Filters are a list of strings (or numbers), and the field must be
        equal to one of them for the row to end up in the result set.

        Note: No filter (filters=None, or filters=[]) means no restriction.
        As it doesn't really make sense to remove all values (and thus all rows)
        from the resulting data, we treat both as "don't filter" in this case
        """
        if not filter_values:
            return

        filters_sql = ", ".join(
            [self.makeparam(val, f"flt_{alias}") for val in filter_values]
        )
        self.filters.append(f"{alias} IN ({filters_sql})")

    @classmethod
    def from_queryset(cls, queryset):
        """Take a Django queryset and represent it as a Query object.

        We will use this via CTE (WITH) statement to read data respecting
        the visibility rules as defined in Caluma.
        """

        # The queryset may "SELECT" columns (especially from JOINs) that
        # we're not interested in, as we're JOINing our stuff by ourselves.
        # To avoid ambiguity, we select ONLY what's in the main model's
        # field list
        q = connection.ops.quote_name
        field_list = ",  \n".join(
            [
                q(queryset.model._meta.db_table)
                + "."
                + q(field.db_column or field.attname)
                for field in queryset.model._meta.concrete_fields
            ]
        )

        # Unfortunately, queryset.query.sql_with_params() seems to
        # ignore queryset.only() specification, so we must do it ourselves.
        # Ugly AF, but don't see how else to do it for now...
        sql, params = queryset.query.sql_with_params()
        select, from_part = sql.split("FROM", 1)

        wrapped_sql = f"""SELECT {field_list}\nFROM\n      {from_part} -- qs ref\n"""

        name = _make_name(value=wrapped_sql, name_hint=queryset.model.__name__.lower())
        self = cls(from_=name, is_plain_cte=True)

        # Django does positional parameters, we do named params. Convert them
        named_params = [self.makeparam(param) for param in params]
        final_sql = wrapped_sql % named_params

        self.with_queries[name] = final_sql

        return self


@dataclass
class Field:
    identifier: str
    alias: Optional[str] = field(default=None)
    parent: Optional[Field] = field(default=None)
    filter_values: Optional[List[str]] = field(default=None)
    answer_value_mode: bool = field(default=False)
    _activated_query: Optional[Query] = field(default=None)

    def path_from_root(self):
        path = []
        obj = self
        while obj:
            path.append(obj)
            obj = obj.parent
        return list(reversed(path))

    def _alias(self):
        if not self.alias:  # pragma: todo cover
            self.alias = slugify(self.identifier + "_" + str(uuid4())[:5]).replace(
                "-", "_"
            )
        return self.alias


class NOOPField(Field):
    def annotate(self, query: Query):
        return query


@dataclass
class AttrField(Field):
    def annotate(self, query: Query):
        if self._activated_query:  # pragma: no cover
            assert query is self._activated_query
            return query

        self._activated_query = query
        query.select.append((self.expr(query), self._alias()))
        return query

    def expr(self, query):
        # TODO: should the table alias be prefixed in all field expr()s?
        alias = query.self_alias()
        if self.answer_value_mode:
            # Caluma form answers are in a JSON field named "value", from where
            # we need to extract the actual value, otherwise, for strings, we
            # would get the quoted version back
            extractor_op = "#>>'{}'"  # noqa
            return f"({alias}.{self.identifier} {extractor_op})"
        return f"{alias}.{self.identifier}"


@dataclass
class DateExprField(AttrField):
    extract_part: Optional[str] = field(default=None)

    def expr(self, query):
        q_id = connection.ops.quote_name(self.identifier)
        return f"EXTRACT({self.extract_part} FROM {q_id})"


@dataclass
class JSONExtractorField(AttrField):
    json_key: Optional[str] = field(default=None)

    def expr(self, query):
        key_param = query.makeparam(self.json_key)
        q_id = connection.ops.quote_name(self.identifier)
        self_alias = query.self_alias()
        # Extract text from JSON field, so that it comes from the DB
        # as actual text
        extractor_op = "#>>'{}'"  # noqa
        return f"""(({self_alias}.{q_id} -> {key_param}) {extractor_op})"""


@dataclass
class JoinField(Field):
    table: Optional[Union[str, Query]] = field(default=None)
    filters: List[str] = field(default_factory=list)
    order_by: List[str] = field(default_factory=list)
    outer_ref: Optional[Tuple[str, str]] = field(
        default=None
    )  # name_in_outer -> name_in_inner

    def annotate(self, query: Query):
        if self._activated_query:  # pragma: no cover (safeguard)
            return self._activated_query

        self._activated_query = Query(
            from_=self.table,
            # distinct_on is the inner part of the outer_ref
            distinct_on=self.outer_ref[1],
            filters=self.filters,
            outer_ref=self.outer_ref,
            order_by=self.order_by,
            # limit=1,
        )

        outer_ref, inner_ref = self.outer_ref

        query.joins.append(
            (
                self._activated_query,
                f'{query.self_alias()}.{outer_ref} = "{self._activated_query.outer_alias()}".{inner_ref}',
            )
        )
        return self._activated_query


class QueryRender:
    """Render a Query structure into actual SQL.

    The Renderer builds an SQL query, which might consist of multiple
    subqueries, and common table expressions (CTE; "WITH" statements).
    """

    def __init__(self, query: Query, is_subquery=False, indent: int = 0):
        self._indent = indent
        self.query = query
        self.collected_params = {}
        self.with_queries = {}
        self.is_subquery = is_subquery

    def as_sql(self, alias):
        """Return a two-tuple: An SQL statement, and a dict of parameters.

        The parameters should then be passed on to the DB driver, which will
        interpret them accordingly.
        """
        self.collected_params = self.query.params.copy()
        self.with_queries = self.query.with_queries.copy()

        field_list = self._field_list()
        if not field_list and self.query.is_plain_cte:
            return self.query.from_, {}, alias

        selects = f"SELECT {self._distinct_on()}\n       {field_list}"
        from_list = f"FROM {self._from_list()}"
        if self.is_subquery:
            sql = "\n".join(
                [
                    selects,
                    from_list,
                    self._where_list(),
                    self._order_by(),
                    self._limits(),
                ]
            )
        else:
            # We're the outermost query. Wrap the whole thing in one more level of
            # subquery, so we can filter on the aliases instead of expressions.
            # This is theoretically only required if we have expressions on the
            # outermost table, but for convenience, we do it always.
            inner_sql = "\n".join(
                [selects, from_list, self._order_by(), self._limits()]
            )
            inner_name = _make_name(inner_sql, "analytics")
            sql = f"SELECT * FROM ({inner_sql}) AS {inner_name}\n{self._where_list()}"

        if self.with_queries and not self.is_subquery:
            # CTEs are only built if we're the outermost query, and
            # we do have CTEs defined
            sql = f"WITH\n{self._with_sql()}\n{sql}"

        return (
            textwrap.indent(sql, prefix=" " * self._indent),
            self.collected_params,
            alias,
        )

    def _with_sql(self):
        with_pairs = [
            textwrap.indent(f"{name} AS ({query})", prefix=" " * (self._indent + 4))
            for name, query in self.with_queries.items()
        ]
        return ",\n".join(with_pairs)

    def _distinct_on(self):
        # DISTINCT ON(foo) only takes the first "foo" row. We use this
        # for example when selecting one sub-item per main table, where
        # multiple might exist (such as the first workitem on a case)
        if self.query.distinct_on:
            return f"DISTINCT ON ({self.query.distinct_on})"
        return ""

    def _order_by(self):
        order_list = []
        if self.query.distinct_on:
            order_list.append(self.query.distinct_on)
        if self.query.order_by:
            # TODO: make parameterizable
            order_list.extend(self.query.order_by)

        if order_list:
            order_sql = ",\n       ".join(order_list)
            return f"ORDER BY {order_sql}"
        return ""

    def _limits(self):
        # TODO: Do we still need this? We're doing the same thing now
        # using (DISTINCT ON)
        if self.query.limit:  # pragma: no cover
            return f"LIMIT {self.query.limit}"
        return ""

    def _where_list(self):
        if self.query.filters:
            # TODO: make parameterizable
            filter_sql = " AND\n       ".join(self.query.filters)
            return f"WHERE {filter_sql}"
        return ""

    def _field_list(self):
        # we always need to select everything, even from subqueries
        fields = [
            # First, the direct fields
            f'''{expr} AS "{alias}"'''
            for expr, alias in self.query.select
        ]
        if self.query.outer_ref:
            _, inner = self.query.outer_ref
            fields.append(inner)

        # The subquery fields are already aliased, so no need
        # to re-alias them
        def _collect(q):
            for _, alias in q.select:
                yield alias
            for sub_q, _ in q.joins:
                yield from _collect(sub_q)

        for join, _ in self.query.joins:
            fields.extend(_collect(join))

        return ",\n       ".join(fields)

    def _join_source(self, from_spec):
        """Return a 2-tuple to join the given table/subquery.

        The tuple consists of:
        * subquery/table
        * alias
        """
        if isinstance(from_spec, str):
            # by defintion, only "our own" table can be passed as string.
            # subqueries MUST be query objects
            return (from_spec, {}, self.query.outer_alias())

        render = QueryRender(from_spec, indent=self._indent + 4, is_subquery=True)

        sql, params, alias = render.as_sql(from_spec.outer_alias())

        # Take any CTEs from the subquery and apply them to "us"
        self.with_queries.update(render.with_queries)

        return sql, params, alias

    def _from_list(self):

        from_sql, from_params, from_alias = self._join_source(self.query.from_)
        self.collected_params.update(from_params)
        self.with_queries.update(self.query.with_queries)

        sources = [(from_sql, from_alias, "")]

        for tbl, cond in self.query.joins:
            subquery, subparams, alias = self._join_source(tbl)
            self.collected_params.update(subparams)
            if isinstance(tbl, Query):
                self.with_queries.update(tbl.with_queries)

            sources.append(
                (f"LEFT JOIN (\n{subquery}\n)", alias, f"ON ({cond})" if cond else "")
            )

        return "\n".join([f"{tbl} AS {alias} {cond}" for tbl, alias, cond in sources])



from django.core import exceptions
from django.db import transaction

from ..caluma_core import serializers
from . import models


class StartingObjectField(serializers.CalumaChoiceField):
    def __init__(self, **kwargs):
        super().__init__(models.AnalyticsTable.STARTING_OBJECT_CHOICES, **kwargs)


class SaveAnalyticsTableSerializer(serializers.ModelSerializer):
    starting_object = StartingObjectField(required=True)

    class Meta:
        model = models.AnalyticsTable
        fields = [
            "slug",
            "name",
            "starting_object",
            "created_at",
            "modified_at",
            "disable_visibilities",
            "meta",
            "created_by_user",
            "created_by_group",
            "modified_by_user",
            "modified_by_group",
        ]


class RemoveAnalyticsTableSerializer(serializers.ModelSerializer):
    slug = serializers.GlobalIDField()

    @transaction.atomic
    def update(self, instance, validated_data):
        instance.delete()
        return instance

    class Meta:
        lookup_input_kwarg = "slug"
        fields = ["slug"]
        model = models.AnalyticsTable


class SaveAnalyticsFieldSerializer(serializers.ModelSerializer):
    id = serializers.GlobalIDField(required=False)

    def validate(self, data):
        validated_data = super().validate(data)
        self._validate_uniqueness(validated_data)
        self._validate_field_exists(validated_data)
        return validated_data

    def _validate_field_exists(self, data):
        start = data["table"].get_starting_object(self.context["info"])
        data_source = data["data_source"]
        try:
            start.get_field(data_source.split("."))
        except KeyError:
            raise exceptions.ValidationError(
                {
                    "dataSource": [
                        f"Specified data source '{data_source}' is not available"
                    ]
                }
            )

    def _validate_uniqueness(self, data):
        existing_fields = data["table"].fields.all()
        if self.instance:
            existing_fields = existing_fields.exclude(pk=self.instance.pk)

        if existing_fields.filter(alias=data["alias"]).exists():
            raise exceptions.ValidationError(
                {"alias": ["Cannot use the same alias twice within the same table"]}
            )
        if existing_fields.filter(data_source=data["data_source"]).exists():
            raise exceptions.ValidationError(
                {
                    "dataSource": [
                        "Cannot use the same data source twice within the same table"
                    ]
                }
            )

    class Meta:
        model = models.AnalyticsField
        fields = [
            "id",
            "alias",
            "table",
            "data_source",
            "created_at",
            "modified_at",
            "filters",
            "meta",
            "created_by_user",
            "created_by_group",
            "modified_by_user",
            "modified_by_group",
        ]


class RemoveAnalyticsFieldSerializer(serializers.ModelSerializer):
    id = serializers.GlobalIDField()

    @transaction.atomic
    def update(self, instance, validated_data):
        instance.delete()

        return instance

    class Meta:
        fields = ["id"]
        model = models.AnalyticsField



# Generated by Django 2.2.10 on 2020-02-27 13:42

from django.db import migrations

from caluma.utils import fix_foreign_key_types


def fix_fk_types_migration(apps, schema_editor):

    fix_foreign_key_types(apps, schema_editor.connection)


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_form", "0033_slugfield_length"),
        ("caluma_workflow", "0019_slugfield_length"),
    ]

    operations = [
        migrations.RunPython(fix_fk_types_migration, migrations.RunPython.noop)
    ]



# Generated by Django 2.2.24 on 2022-01-10 10:51

from django.db import migrations
import localized_fields.fields.field


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_form", "0041_action_button_question"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalquestion",
            name="hint_text",
            field=localized_fields.fields.field.LocalizedField(
                blank=True, null=True, required=[]
            ),
        ),
        migrations.AddField(
            model_name="question",
            name="hint_text",
            field=localized_fields.fields.field.LocalizedField(
                blank=True, null=True, required=[]
            ),
        ),
    ]



import json
from collections import defaultdict

from django.core.management.base import BaseCommand

from caluma.caluma_analytics.models import AnalyticsTable
from caluma.caluma_analytics.simple_table import SimpleTable


class Command(BaseCommand):
    help = "Run an analytics table and show it's output"

    def add_arguments(self, parser):
        parser.add_argument("id", nargs="*", help="Identifier of the analytics table")
        parser.add_argument(
            "--json", action="store_true", help="Request output in JSON format"
        )
        parser.add_argument(
            "--sql",
            action="store_true",
            help="Instead of outputting the analysis, show the SQL",
        )
        parser.add_argument(
            "--sqlonly",
            action="store_true",
            help="Only output SQL, do not run the analysis",
        )

    def show_existing_tables(self):
        print("No analytics table specified. The following tables are available:")
        print("  ID:                  NAME:")
        for table in AnalyticsTable.objects.all():
            print(f"  {table.slug:20} {table.name}")

    def handle(self, **options):
        if not len(options["id"]):
            return self.show_existing_tables()

        analytics_table = AnalyticsTable.objects.get(pk__in=options["id"])
        table = SimpleTable(analytics_table)

        if options["sqlonly"]:
            self.show_sql(table)
            return
        if options["sql"]:
            self.show_sql(table)

        try:
            records = table.get_records()

            if not records:  # pragma: no cover
                return

            if options["json"]:
                self.show_json(records)
            else:
                self.show_table(records)
        except (BrokenPipeError, KeyboardInterrupt):  # pragma: no cover
            # if user presses Ctrl+C, or runs output into
            # a pipe and stops that, we don't bother telling
            # them what happened
            pass

    def show_sql(self, table):
        # Writing SQL dump to STDERR, so users can still
        # split/pipe data output to somewhere else
        sql, params = table.get_sql_and_params()
        self.stderr.write("-- SQL: \n")
        self.stderr.write(sql)
        self.stderr.write("-- PARAMS: \n")
        for name, val in params.items():
            self.stderr.write(f"--     {name}: {val}\n")
        self.stderr.flush()

    def show_json(self, records):
        print(
            json.dumps(
                [{k: str(v) for k, v in rec.items()} for rec in records], indent=4
            )
        )

    def show_table(self, records):
        """Output the analytics table as an ASCII table to the console."""

        col_lengths = defaultdict(int)

        for rec in records:
            for key, val in rec.items():
                new_len = max(col_lengths[key], len(str(val)), len(key))
                col_lengths[key] = new_len

        format_string = " ".join(
            [
                "{" + col + ":<" + str(length + 2) + "}"
                for col, length in col_lengths.items()
            ]
        )
        print(format_string.format(**{k: k for k in col_lengths}))
        print(format_string.format(**{k: "-" * l for k, l in col_lengths.items()}))

        for rec in records:
            print(format_string.format(**{k: str(v) for k, v in rec.items()}))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.17 on 2018-12-28 12:43
from __future__ import unicode_literals

import django.contrib.postgres.fields
import django.contrib.postgres.indexes
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0004_auto_20181218_1352")]

    operations = [
        migrations.AddField(
            model_name="task",
            name="address_groups",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="workitem",
            name="addressed_groups",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=150),
                default=list,
                help_text="Offer work item to be processed by a group of users, such are not committed to process it though.",
                size=None,
            ),
        ),
        migrations.AddField(
            model_name="workitem",
            name="assigned_users",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=150),
                default=list,
                help_text="Users responsible to undertake given work item.",
                size=None,
            ),
        ),
        migrations.AlterField(
            model_name="case",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="case",
            name="meta",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="flow",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="meta",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="taskflow",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="meta",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="workitem",
            name="child_case",
            field=models.OneToOneField(
                blank=True,
                help_text="Defines case of a sub-workflow",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="parent_work_item",
                to="caluma_workflow.Case",
            ),
        ),
        migrations.AlterField(
            model_name="workitem",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="workitem",
            name="meta",
            field=models.JSONField(default=dict),
        ),
        migrations.AddIndex(
            model_name="workitem",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["addressed_groups"], name="workflow_wo_address_679262_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="workitem",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["assigned_users"], name="workflow_wo_assigne_76d859_gin"
            ),
        ),
    ]



import re
from collections import namedtuple

from django.conf import settings
from django.utils.module_loading import import_string
from localized_fields.value import LocalizedValue
from rest_framework.exceptions import ValidationError


class BaseFormatValidator:
    r"""Basic format validator class to be extended by any format validator implementation.

    A custom format validator class could look like this:
    ```
    >>> from caluma.caluma_form.format_validators import BaseFormatValidator
    ...
    ...
    ... class CustomFormatValidator(BaseFormatValidator):
    ...     slug = "email"
    ...     name = {"en": "E-mail", "de": "E-Mail"}
    ...     regex = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)"
    ...     error_msg = {"en": "Not an e-mail address", "de": "Keine E-Mail adresse"}
    ```
    """

    # `name` and `error_msg` are just dicts and get cast to `LocalizedValue` when needed.
    #  This is to not break with the existing API.

    def __init__(self):
        if not all(
            [self.slug, self.regex, self.name, self.error_msg]
        ):  # pragma: no cover
            raise NotImplementedError("Missing properties!")

    def validate(self, value, document):
        if not re.match(self.regex, value):
            raise ValidationError(LocalizedValue(self.error_msg).translate())


class EMailFormatValidator(BaseFormatValidator):
    slug = "email"
    name = {"en": "E-mail address", "de": "E-Mail Adresse", "fr": "Adresse e-mail"}
    regex = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)"
    error_msg = {
        "en": "Please enter a valid e-mail address",
        "de": "Bitte geben Sie eine gültige E-Mail Adresse ein",
        "fr": "Veuillez entrer une addresse e-mail valide",
    }


class PhoneNumberFormatValidator(BaseFormatValidator):
    slug = "phone-number"
    name = {"en": "Phone number", "de": "Telefonnummer", "fr": "Numéro de téléphone"}
    regex = r"^[\s\/\.\(\)-]*(?:\+|0|00)(?:[\s\/\.\(\)-]*\d[\s\/\.\(\)-]*){6,20}$"
    error_msg = {
        "en": "Please enter a valid phone number",
        "de": "Bitte geben Sie eine gültige Telefonnummer ein",
        "fr": "Veuillez entrer un numéro de téléphone valide",
    }


base_format_validators = [EMailFormatValidator, PhoneNumberFormatValidator]


FormatValidator = namedtuple("FormatValidator", ["slug", "name", "regex", "error_msg"])


def get_format_validators(include=None, dic=False):
    """Get all FormatValidators.

    :param include: List of FormatValidators to include
    :param dic: Should return a dict
    :return: List of FormatValidator-objects if dic False otherwise dict
    """

    format_validator_classes = [
        import_string(cls) for cls in settings.FORMAT_VALIDATOR_CLASSES
    ] + base_format_validators
    if include is not None:
        format_validator_classes = [
            fvc for fvc in format_validator_classes if fvc.slug in include
        ]
    if dic:
        return {ds.slug: ds for ds in format_validator_classes}
    return [
        FormatValidator(
            slug=ds.slug,
            name=LocalizedValue(ds.name).translate(),
            regex=ds.regex,
            error_msg=LocalizedValue(ds.error_msg).translate(),
        )
        for ds in format_validator_classes
    ]



# Generated by Django 2.2.10 on 2020-02-26 09:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0032_auto_20200220_1311")]

    operations = [
        migrations.AlterField(
            model_name="form",
            name="slug",
            field=models.SlugField(max_length=127, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="historicalform",
            name="slug",
            field=models.SlugField(max_length=127),
        ),
        migrations.AlterField(
            model_name="historicaloption",
            name="slug",
            field=models.SlugField(max_length=127),
        ),
        migrations.AlterField(
            model_name="historicalquestion",
            name="slug",
            field=models.SlugField(max_length=127),
        ),
        migrations.AlterField(
            model_name="option",
            name="slug",
            field=models.SlugField(max_length=127, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="question",
            name="slug",
            field=models.SlugField(max_length=127, primary_key=True, serialize=False),
        ),
    ]



# Generated by Django 3.2.12 on 2022-03-09 11:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_workflow", "0028_redoable_field"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicaltask",
            name="continue_async",
            field=models.BooleanField(
                default=False,
                help_text="Whether to continue the flow if the multiple instance work item has ready siblings",
            ),
        ),
        migrations.AddField(
            model_name="task",
            name="continue_async",
            field=models.BooleanField(
                default=False,
                help_text="Whether to continue the flow if the multiple instance work item has ready siblings",
            ),
        ),
    ]



from caluma.caluma_form import models as form_models, schema as form_schema
from caluma.caluma_workflow import schema as workflow_schema

from . import sql


def qs_method_factory(node_cls, model_cls=None):
    if not model_cls:
        model_cls = node_cls._meta.model

    def method(self):
        queryset = model_cls.objects.all()
        if not self.is_disabled:
            queryset = node_cls.get_queryset(queryset, self.info)
        return sql.Query.from_queryset(queryset)

    method.__doc__ = f"""
    Return an analytics Query object for {model_cls.__name__}.

    The query object contains CTE representing the visibility rules
    as defined in Caluma's settings for {model_cls.__name__}.
    """

    return method


class CalumaVisibilitySource:
    def __init__(self, info, is_disabled):
        self.info = info
        self.is_disabled = is_disabled

    cases = qs_method_factory(workflow_schema.Case)
    workitems = qs_method_factory(workflow_schema.WorkItem)
    documents = qs_method_factory(form_schema.Document)
    answers = qs_method_factory(form_schema.Answer, form_models.Answer)



from graphql_relay import from_global_id

_valid_types = set()


def extract_global_id(id):
    """
    Extract id from global id throwing away type.

    In case it is not a base64 encoded value it will return id as is.
    This way it can be used as input type by simply defining id without its specific type.
    """
    if not _valid_types:
        # Lazy import due to circular dependency
        from caluma import schema
        from caluma.caluma_core import types

        all_types_dict = {
            **vars(schema.workflow_schema),
            **vars(schema.form_schema),
            **vars(schema.form_historical_schema),
            **vars(schema.analytics_schema),
        }
        all_types = [
            typename
            for typename, gql_type in all_types_dict.items()
            if isinstance(gql_type, type) and issubclass(gql_type, types.Node)
        ]

        _valid_types.update(all_types)

    gql_type, result = from_global_id(id)

    # from_global_id() now returns an empty string if the passed id could not be
    # decoded from base64 which is the case for raw ids
    if result == "":
        return id

    # some valid ids could decode to a string that from_global_id()
    # assumes is a global id, so we need to verify that it is indeed
    # the case.
    if gql_type and gql_type not in _valid_types:
        return id

    return result



import graphene
from django.http import Http404
from graphene import relay
from graphene.types import ObjectType, generic

from ..caluma_core.relay import extract_global_id
from ..caluma_core.types import ConnectionField, CountableConnectionBase
from . import models
from .schema import (
    QUESTION_ANSWER_TYPES,
    Answer,
    DateAnswer,
    FileAnswer,
    FloatAnswer,
    FormDjangoObjectType,
    IntegerAnswer,
    ListAnswer,
    StringAnswer,
    TableAnswer,
)
from .storage_clients import client


def historical_qs_as_of(queryset, date, pk_attr):
    """Get history revision as of `date` for queryset.

    :param queryset: history Queryset()
    :param date: aware datetime()
    :param pk_attr: str (pk field name)
    :return: Queryset()
    """
    # This could become unnecessary as soon as
    # https://github.com/treyhunner/django-simple-history/issues/397 is resolved.
    return (
        queryset.filter(history_date__lte=date)
        .order_by(pk_attr, "-history_date")
        .distinct(pk_attr)
    )


class HistoricalAnswer(Answer):
    history_date = graphene.types.datetime.DateTime(required=True)
    history_user_id = graphene.String()
    history_type = graphene.String()

    @classmethod
    def resolve_type(cls, instance, info):
        answer_type = QUESTION_ANSWER_TYPES[instance.history_question_type]
        return f"Historical{answer_type.__name__}"


class HistoricalAnswerConnection(CountableConnectionBase):
    class Meta:
        node = HistoricalAnswer


class HistoricalIntegerAnswer(IntegerAnswer):
    class Meta:
        model = models.Answer.history.model
        exclude = ("document", "file", "date")
        use_connection = False
        interfaces = (HistoricalAnswer, graphene.Node)


class HistoricalFloatAnswer(FloatAnswer):
    class Meta:
        model = models.Answer.history.model
        exclude = ("document", "file", "date")
        use_connection = False
        interfaces = (HistoricalAnswer, graphene.Node)


class HistoricalDateAnswer(DateAnswer):
    class Meta:
        model = models.Answer.history.model
        exclude = ("document", "file")
        use_connection = False
        interfaces = (HistoricalAnswer, graphene.Node)


class HistoricalStringAnswer(StringAnswer):
    class Meta:
        model = models.Answer.history.model
        exclude = ("document", "file", "date")
        use_connection = False
        interfaces = (HistoricalAnswer, graphene.Node)


class HistoricalListAnswer(ListAnswer):
    class Meta:
        model = models.Answer.history.model
        exclude = ("document", "file", "date")
        use_connection = False
        interfaces = (HistoricalAnswer, graphene.Node)


class HistoricalFile(ObjectType):
    name = graphene.String(required=True)
    download_url = graphene.String()
    metadata = generic.GenericScalar()
    historical_answer = graphene.Field(
        "caluma.caluma_form.historical_schema.HistoricalFileAnswer"
    )
    history_date = graphene.types.datetime.DateTime(required=True)
    history_user_id = graphene.String()
    history_type = graphene.String()

    @classmethod
    def resolve_download_url(cls, instance, info):
        # if the file is part of the newest revision, the downloadUri has to be for the
        # actual file object_name.
        if instance == instance.instance.history.first():
            return instance.instance.download_url
        return client.download_url(f"{instance.pk}_{instance.name}")

    class Meta:
        model = models.File.history.model
        interfaces = (relay.Node,)


class HistoricalFileAnswer(FileAnswer):
    value = graphene.Field(
        HistoricalFile,
        required=False,
        as_of=graphene.types.datetime.DateTime(required=True),
    )

    def resolve_value(self, info, as_of, **args):
        # we need to use the HistoricalFile of the correct revision
        return models.File.history.filter(
            id=self.file_id, history_date__lte=as_of
        ).first()

    class Meta:
        model = models.Answer.history.model
        exclude = ("document", "date")
        use_connection = False
        interfaces = (HistoricalAnswer, graphene.Node)


class HistoricalDocument(FormDjangoObjectType):
    historical_answers = ConnectionField(
        HistoricalAnswerConnection,
        as_of=graphene.types.datetime.DateTime(required=True),
    )
    history_date = graphene.types.datetime.DateTime(required=True)
    history_user_id = graphene.String()
    history_type = graphene.String()
    meta = generic.GenericScalar()
    document_id = graphene.UUID()

    def resolve_document_id(self, info, *args, **kwargs):
        return self.id

    def resolve_historical_answers(self, info, as_of, *args, **kwargs):
        return historical_qs_as_of(
            models.Answer.history.filter(document_id=self.id), as_of, "id"
        )

    class Meta:
        model = models.Document.history.model
        exclude = ("family", "history_id", "history_change_reason")
        interfaces = (graphene.Node,)
        connection_class = CountableConnectionBase


class HistoricalTableAnswer(TableAnswer):
    value = graphene.List(
        HistoricalDocument,
        required=False,
        as_of=graphene.types.datetime.DateTime(required=True),
    )

    def resolve_value(self, info, as_of, *args):
        answerdocuments_unordered = historical_qs_as_of(
            models.AnswerDocument.history.filter(answer_id=self.id), as_of, "id"
        )

        # ordering has to happen in a separate query because of the use of `distinct()`
        answerdocuments = models.AnswerDocument.history.filter(
            pk__in=answerdocuments_unordered
        ).order_by("sort")

        documents = [
            models.Document.history.filter(
                id=ad.document_id, history_date__lte=as_of
            ).latest("history_date")
            for ad in answerdocuments
        ]

        # Since python 3.6, `list(dict.fromkeys(somelist))` is the most performant way
        # to remove duplicates from a list, while retaining it's order.
        # In python 3.6 this is an implementation detail. From python 3.7 onwards it is
        # a language feature.
        # Luckily django model instances are hashable, so we're able to make use of
        # this.
        return list(dict.fromkeys(documents))

    class Meta:
        model = models.Answer.history.model
        exclude = ("file", "date")
        use_connection = False
        interfaces = (HistoricalAnswer, graphene.Node)


def document_as_of(info, document_global_id, timestamp):
    document_id = extract_global_id(document_global_id)
    document_qs = HistoricalDocument.get_queryset(models.Document.history.all(), info)
    document = document_qs.filter(id=document_id, history_date__lte=timestamp).first()
    if not document:
        raise Http404("No HistoricalDocument matches the given query.")
    return document


class Query:
    document_as_of = graphene.Field(
        HistoricalDocument,
        id=graphene.ID(required=True),
        as_of=graphene.types.datetime.DateTime(required=True),
    )

    def resolve_document_as_of(self, info, id, as_of):
        return document_as_of(info, id, as_of)



# Generated by Django 3.2.11 on 2022-02-10 09:48

import uuid

import django.contrib.postgres.fields
import django.db.models.deletion
import localized_fields.fields.field
import simple_history.models
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AnalyticsTable",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "created_by_user",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "modified_by_user",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "modified_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "slug",
                    models.SlugField(max_length=127, primary_key=True, serialize=False),
                ),
                ("meta", models.JSONField(default=dict)),
                ("disable_visibilities", models.BooleanField(default=False)),
                ("name", localized_fields.fields.field.LocalizedField(required=[])),
                (
                    "starting_object",
                    models.CharField(choices=[("cases", "Cases")], max_length=250),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="HistoricalAnalyticsTable",
            fields=[
                ("history_user_id", models.CharField(max_length=150, null=True)),
                (
                    "created_at",
                    models.DateTimeField(blank=True, db_index=True, editable=False),
                ),
                (
                    "modified_at",
                    models.DateTimeField(blank=True, db_index=True, editable=False),
                ),
                (
                    "created_by_user",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "modified_by_user",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "modified_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("slug", models.SlugField(max_length=127)),
                ("meta", models.JSONField(default=dict)),
                ("disable_visibilities", models.BooleanField(default=False)),
                ("name", localized_fields.fields.field.LocalizedField(required=[])),
                (
                    "starting_object",
                    models.CharField(choices=[("cases", "Cases")], max_length=250),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical analytics table",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalAnalyticsField",
            fields=[
                ("history_user_id", models.CharField(max_length=150, null=True)),
                (
                    "created_at",
                    models.DateTimeField(blank=True, db_index=True, editable=False),
                ),
                (
                    "modified_at",
                    models.DateTimeField(blank=True, db_index=True, editable=False),
                ),
                (
                    "created_by_user",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "modified_by_user",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "modified_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "id",
                    models.UUIDField(db_index=True, default=uuid.uuid4, editable=False),
                ),
                ("alias", models.CharField(max_length=100)),
                ("meta", models.JSONField(default=dict)),
                ("data_source", models.TextField()),
                (
                    "filters",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.TextField(),
                        blank=True,
                        default=list,
                        null=True,
                        size=None,
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "table",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_analytics.analyticstable",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical analytics field",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="AnalyticsField",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "created_by_user",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "modified_by_user",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "modified_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("alias", models.CharField(max_length=100)),
                ("meta", models.JSONField(default=dict)),
                ("data_source", models.TextField()),
                (
                    "filters",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.TextField(),
                        blank=True,
                        default=list,
                        null=True,
                        size=None,
                    ),
                ),
                (
                    "table",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fields",
                        to="caluma_analytics.analyticstable",
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="analyticsfield",
            constraint=models.UniqueConstraint(
                fields=("table", "data_source"), name="unique_data_source"
            ),
        ),
        migrations.AddConstraint(
            model_name="analyticsfield",
            constraint=models.UniqueConstraint(
                fields=("table", "alias"), name="unique_alias"
            ),
        ),
    ]



from ..caluma_core.permissions import BasePermission


class IsAuthenticated(BasePermission):
    """Only allow authenticated users to execute mutations."""

    def has_permission(self, mutation, info):
        return info.context.user.is_authenticated


class CreatedByGroup(BasePermission):
    """Only allow mutating data that belongs to same group as current user."""

    def has_object_permission(self, mutation, info, instance):
        return instance.created_by_group in info.context.user.groups



from django.db import migrations


def get_validators(question):
    return list(
        set([*question.meta.get("formatValidators", []), *question.format_validators])
    )


def forward(apps, schema_editor):
    Question = apps.get_model("caluma_form", "Question")

    for question in Question.objects.filter(meta__formatValidators__isnull=False):
        question.format_validators = get_validators(question)
        del question.meta["formatValidators"]
        question.save()


def reverse(apps, schema_editor):
    Question = apps.get_model("caluma_form", "Question")

    for question in Question.objects.exclude(format_validators=[]):
        question.meta.update({"formatValidators": get_validators(question)})
        question.format_validators = []
        question.save()


class Migration(migrations.Migration):
    dependencies = [
        ("caluma_form", "0043_alter_historicalanswer_history_question_type")
    ]

    operations = [migrations.RunPython(forward, reverse)]



from functools import reduce

import graphene
from django.core import exceptions
from django.db import ProgrammingError
from django.db.models import Q
from django.forms import BooleanField
from django.utils import translation
from django_filters.constants import EMPTY_VALUES
from django_filters.rest_framework import CharFilter, Filter, FilterSet
from graphene import Enum, InputObjectType, List
from graphene_django.forms.converter import convert_form_field
from graphene_django.registry import get_global_registry

from ..caluma_core.filters import (
    CompositeFieldClass,
    GlobalIDFilter,
    GlobalIDMultipleChoiceFilter,
    MetaFilterSet,
    OrderingFilter,
    SearchFilter,
    SlugMultipleChoiceFilter,
)
from ..caluma_core.ordering import AttributeOrderingFactory, MetaFieldOrdering
from ..caluma_form.models import Answer, DynamicOption, Question
from ..caluma_form.ordering import AnswerValueOrdering
from . import models, validators


class FormFilterSet(MetaFilterSet):
    search = SearchFilter(fields=("slug", "name", "description"))
    order_by = OrderingFilter(label="FormOrdering", fields=("name",))
    slugs = SlugMultipleChoiceFilter(field_name="slug")
    slug = CharFilter()
    slug.deprecation_reason = "Use the `slugs` (plural) filter instead, which allows filtering for multiple slugs"
    questions = SlugMultipleChoiceFilter(field_name="questions__slug")

    class Meta:
        model = models.Form
        fields = (
            "slug",
            "name",
            "description",
            "is_published",
            "is_archived",
            "questions",
        )


class FormOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.Form,
        fields=[
            "created_at",
            "modified_at",
            "created_by_user",
            "created_by_group",
            "modified_by_user",
            "modified_by_group",
            "slug",
            "name",
            "description",
            "is_published",
            "is_archived",
        ],
    )

    class Meta:
        model = models.Form
        fields = ("meta", "attribute")


class OptionFilterSet(MetaFilterSet):
    search = SearchFilter(fields=("slug", "label"))
    order_by = OrderingFilter(label="OptionOrdering", fields=("label",))

    class Meta:
        model = models.Option
        fields = ("slug", "label", "is_archived")


class QuestionFilterSet(MetaFilterSet):
    exclude_forms = GlobalIDMultipleChoiceFilter(field_name="forms", exclude=True)
    search = SearchFilter(fields=("slug", "label"))
    order_by = OrderingFilter(label="QuestionOrdering", fields=("label",))
    slugs = SlugMultipleChoiceFilter(field_name="slug")
    slug = CharFilter()
    slug.deprecation_reason = "Use the `slugs` (plural) filter instead, which allows filtering for multiple slugs"

    class Meta:
        model = models.Question
        fields = (
            "slug",
            "label",
            "is_required",
            "is_hidden",
            "is_archived",
            "sub_form",
            "row_form",
        )


class QuestionOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.Question,
        exclude_fields=[
            "configuration",
            "data_source",
            "format_validators",
            "meta",
            "row_form",
            "source",
            "static_content",
            "sub_form",
            "default_answer",
            "calc_dependents",
        ],
    )

    class Meta:
        model = models.Question
        fields = ("meta", "attribute")


class AnswerLookupMode(Enum):
    EXACT = "exact"
    STARTSWITH = "startswith"
    CONTAINS = "contains"
    ICONTAINS = "icontains"
    INTERSECTS = "intersects"
    ISNULL = "isnull"
    IN = "in"

    GTE = "gte"
    GT = "gt"
    LTE = "lte"
    LT = "lt"


class AnswerHierarchyMode(Enum):
    DIRECT = "DIRECT"
    FAMILY = "FAMILY"


class HasAnswerFilterType(InputObjectType):
    """
    Lookup type to search document structures.

    When using lookup `ISNULL`, the provided `value` will be ignored.
    """

    question = graphene.ID(required=True)
    value = graphene.types.generic.GenericScalar(required=False)
    lookup = AnswerLookupMode()
    hierarchy = AnswerHierarchyMode()


class HasAnswerFilterField(CompositeFieldClass):
    pass


class HasAnswerFilter(Filter):
    field_class = HasAnswerFilterField

    def __init__(self, *args, **kwargs):
        self.document_id = kwargs.pop("document_id")
        super().__init__(*args, **kwargs)

    VALID_LOOKUPS = {
        "text": [
            AnswerLookupMode.EXACT,
            AnswerLookupMode.STARTSWITH,
            AnswerLookupMode.CONTAINS,
            AnswerLookupMode.ICONTAINS,
            AnswerLookupMode.ISNULL,
            AnswerLookupMode.IN,
        ],
        "integer": [
            AnswerLookupMode.EXACT,
            AnswerLookupMode.LT,
            AnswerLookupMode.LTE,
            AnswerLookupMode.GT,
            AnswerLookupMode.GTE,
            AnswerLookupMode.ISNULL,
            AnswerLookupMode.IN,
        ],
        "choice": [
            AnswerLookupMode.EXACT,
            AnswerLookupMode.IN,
            AnswerLookupMode.ISNULL,
        ],
        "multiple_choice": [
            AnswerLookupMode.EXACT,
            AnswerLookupMode.CONTAINS,
            AnswerLookupMode.INTERSECTS,
            AnswerLookupMode.ISNULL,
        ],
    }
    VALID_LOOKUPS["date"] = VALID_LOOKUPS["integer"]
    VALID_LOOKUPS["float"] = VALID_LOOKUPS["integer"]
    VALID_LOOKUPS["dynamic_choice"] = VALID_LOOKUPS["choice"]
    VALID_LOOKUPS["dynamic_multiple_choice"] = VALID_LOOKUPS["multiple_choice"]
    VALID_LOOKUPS["textarea"] = VALID_LOOKUPS["text"]
    VALID_LOOKUPS["datetime"] = VALID_LOOKUPS["integer"]
    VALID_LOOKUPS["calculated_float"] = VALID_LOOKUPS["float"]

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        for expr in value:
            qs = self.apply_expr(qs, expr)
        return qs

    def apply_expr(self, qs, expr):
        lookup = expr.get("lookup", self.lookup_expr)
        lookup_expr = (hasattr(lookup, "value") and lookup.value) or lookup

        question_slug = expr["question"]
        match_value = expr.get("value")
        if lookup == AnswerLookupMode.ISNULL:
            match_value = True

        hierarchy = expr.get("hierarchy", AnswerHierarchyMode.FAMILY)

        question = models.Question.objects.get(slug=question_slug)
        self._validate_lookup(question, lookup)

        answer_value = "value"
        if question.type == models.Question.TYPE_DATE:
            answer_value = "date"

        answers = models.Answer.objects.all()

        if lookup == AnswerLookupMode.INTERSECTS:
            inner_lookup = "exact"
            if question.type in (
                models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
                models.Question.TYPE_MULTIPLE_CHOICE,
            ):
                inner_lookup = "contains"

            exprs = [
                Q(**{f"value__{inner_lookup}": val, "question__slug": question_slug})
                for val in match_value
            ]
            # connect all expressions with OR
            answers = answers.filter(reduce(lambda a, b: a | b, exprs))
        else:
            answers = answers.filter(
                **{
                    f"{answer_value}__{lookup_expr}": match_value,
                    "question__slug": question_slug,
                }
            )

        if hierarchy == AnswerHierarchyMode.FAMILY:
            return qs.filter(
                **{f"{self.document_id}__in": answers.values("document__family")}
            )
        else:
            return qs.filter(**{f"{self.document_id}__in": answers.values("document")})

    def _validate_lookup(self, question, lookup):
        try:
            valid_lookups = self.VALID_LOOKUPS[question.type]
        except KeyError:  # pragma: no cover
            # Not covered in tests - this only happens when you add a new
            # question type and forget to update the lookup config above.  In
            # that case, the fix is simple - go up a few lines and adjust the
            # VALID_LOOKUPS dict.
            raise ProgrammingError(
                f"Valid lookups not configured for question type {question.type}"
            )

        if lookup not in valid_lookups:
            lookup = (hasattr(lookup, "value") and lookup.value) or lookup
            raise exceptions.ValidationError(
                f"Invalid lookup for question slug={question.slug} ({question.type.upper()}): {lookup.upper()}"
            )

    @staticmethod
    @convert_form_field.register(HasAnswerFilterField)
    def convert_meta_value_field(field):
        registry = get_global_registry()
        converted = registry.get_converted_field(field)
        if converted:
            return converted

        # the converted type must be list-of-filter, as we need to apply
        # multiple conditions
        converted = List(HasAnswerFilterType)
        registry.register_converted_field(field, converted)
        return converted


class SearchLookupMode(Enum):
    STARTSWITH = "startswith"
    CONTAINS = "icontains"
    TEXT = "search"


class SearchAnswersFilterType(InputObjectType):
    """Lookup type to search in answers."""

    questions = List(graphene.ID)
    value = graphene.types.generic.GenericScalar(required=True)
    lookup = SearchLookupMode(required=False)


class SearchAnswersFilterField(CompositeFieldClass):
    pass


class SearchAnswersFilter(Filter):
    field_class = SearchAnswersFilterField

    FIELD_MAP = {
        Question.TYPE_TEXT: "value",
        Question.TYPE_TEXTAREA: "value",
        Question.TYPE_DATE: "date",
        Question.TYPE_CHOICE: "value",
        Question.TYPE_MULTIPLE_CHOICE: "value",
        Question.TYPE_DYNAMIC_CHOICE: "value",
        Question.TYPE_DYNAMIC_MULTIPLE_CHOICE: "value",
        Question.TYPE_INTEGER: "value",
        Question.TYPE_FLOAT: "value",
    }

    def __init__(self, *args, **kwargs):
        self.document_id = kwargs.pop("document_id")
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        assert isinstance(value, list)

        for val in value:
            if val in EMPTY_VALUES:  # pragma: no cover
                continue
            qs = self._apply_filter(qs, val)
        return qs

    def _apply_filter(self, qs, value):

        questions = self._validate_and_get_questions(value["questions"])

        for word in value["value"].split():
            answers_with_word = self._answers_with_word(
                questions, word, value.get("lookup", SearchLookupMode.CONTAINS.value)
            )
            qs = qs.filter(
                **{
                    f"{self.document_id}__in": answers_with_word.values(
                        "document__family"
                    )
                }
            )

        return qs

    def _word_lookup_for_question(self, question, word, lookup):
        if question.type not in (
            Question.TYPE_CHOICE,
            Question.TYPE_DYNAMIC_CHOICE,
            Question.TYPE_MULTIPLE_CHOICE,
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
        ):
            return Q(
                **{
                    f"{self.FIELD_MAP[question.type]}__{lookup}": word,
                    "question": question,
                }
            )

        # (Multiple) choice lookups need more specific treatment...
        lang = translation.get_language()
        is_multiple = question.type in (
            Question.TYPE_MULTIPLE_CHOICE,
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
        )
        is_dynamic = question.type in (
            Question.TYPE_DYNAMIC_CHOICE,
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
        )

        # find all options of our given question that match the
        # word, then use their slugs for lookup
        if is_dynamic:
            matching_options = (
                DynamicOption.objects.all()
                .filter(question=question, **{f"label__{lang}__{lookup}": word})
                .values_list("slug", flat=True)
            )
        else:
            matching_options = question.options.filter(
                **{f"label__{lang}__{lookup}": word}
            ).values_list("slug", flat=True)

        if not matching_options:
            # no labels = no results
            return Q(value=False) & Q(value=True)

        filt = "value__contains" if is_multiple else "value"
        return reduce(
            lambda a, b: a | b, [Q(**{filt: slug}) for slug in matching_options]
        )

    def _answers_with_word(self, questions, word, lookup):
        exprs = [
            self._word_lookup_for_question(question, word, lookup)
            for q_slug, question in questions.items()
        ]

        # join expressions with OR
        answer_qs = Answer.objects.filter(reduce(lambda a, b: a | b, exprs))
        return answer_qs

    def _validate_and_get_questions(self, questions):
        res = {}
        for q_slug in questions:
            question = Question.objects.get(pk=q_slug)
            if question.type not in self.FIELD_MAP:
                raise exceptions.ValidationError(
                    f"Questions of type {question.type} cannot be used in searchAnswers"
                )
            res[q_slug] = question
        return res

    @staticmethod
    @convert_form_field.register(SearchAnswersFilterField)
    def convert_meta_value_field(field):
        registry = get_global_registry()
        converted = registry.get_converted_field(field)
        if converted:
            # TODO: can this be removed, as it's never the case
            # anymore with graphene 3.0?
            # Doesn't seem to have ill effects...
            return converted  # pragma: no cover

        converted = List(SearchAnswersFilterType)
        registry.register_converted_field(field, converted)
        return converted


class DocumentFilterSet(MetaFilterSet):
    id = GlobalIDFilter()
    search = SearchFilter(
        fields=(
            "form__slug",
            "form__name",
            "form__description",
            "answers__value",
            "answers__file__name",
        )
    )
    order_by = OrderingFilter(label="DocumentOrdering")
    root_document = GlobalIDFilter(field_name="family")
    forms = GlobalIDMultipleChoiceFilter(field_name="form")

    has_answer = HasAnswerFilter(document_id="pk")
    search_answers = SearchAnswersFilter(document_id="pk")

    class Meta:
        model = models.Document
        fields = ("form", "forms", "search", "id")


class DocumentOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    answer_value = AnswerValueOrdering()
    attribute = AttributeOrderingFactory(
        models.Document, exclude_fields=["meta", "family", "id"]
    )

    class Meta:
        model = models.Document
        fields = ("meta",)


class VisibleAnswerFilter(Filter):
    field_class = BooleanField

    def filter(self, qs, value):
        if not value or not qs.exists():
            return qs

        # assuming qs can only ever be in the context of a single document
        document = qs.first().document.family
        validator = validators.DocumentValidator()
        return qs.filter(question__slug__in=validator.visible_questions(document))


class AnswerFilterSet(MetaFilterSet):
    search = SearchFilter(fields=("value", "file__name"))
    order_by = OrderingFilter(label="AnswerOrdering")
    questions = GlobalIDMultipleChoiceFilter(field_name="question")

    visible_in_context = VisibleAnswerFilter()

    class Meta:
        model = models.Answer
        fields = ("question", "search")


class AnswerOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(models.Answer, exclude_fields=["meta", "id"])

    class Meta:
        model = models.Answer
        fields = ("meta",)


class DynamicOptionFilterSet(FilterSet):
    question = GlobalIDFilter()
    document = GlobalIDFilter()

    class Meta:
        model = models.DynamicOption
        fields = ("question", "document")



"""
This settings module only contains caluma specific settings.

It's imported by the main caluma settings and is intended to also be used by third party
applications integrating Caluma.
"""


import os

import environ

env = environ.Env()
django_root = environ.Path(__file__) - 3

ENV_FILE = env.str("ENV_FILE", default=django_root(".env"))
if os.path.exists(ENV_FILE):  # pragma: no cover
    environ.Env.read_env(ENV_FILE)

# per default production is enabled for security reasons
# for development create .env file with ENV=development
ENV = env.str("ENV", "production")


def default(default_dev=env.NOTSET, default_prod=env.NOTSET):
    """Environment aware default."""
    return default_prod if ENV == "production" else default_dev


# Managing files

MEDIA_STORAGE_SERVICE = env.str("MEDIA_STORAGE_SERVICE", default="minio")
MINIO_STORAGE_ENDPOINT = env.str("MINIO_STORAGE_ENDPOINT", default="minio:9000")
MINIO_STORAGE_ACCESS_KEY = env.str("MINIO_STORAGE_ACCESS_KEY", default="minio")
MINIO_STORAGE_SECRET_KEY = env.str("MINIO_STORAGE_SECRET_KEY", default="minio123")
MINIO_STORAGE_USE_HTTPS = env.str("MINIO_STORAGE_USE_HTTPS", default=False)
MINIO_STORAGE_MEDIA_BUCKET_NAME = env.str(
    "MINIO_STORAGE_MEDIA_BUCKET_NAME", default="caluma-media"
)
MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET = env.str(
    "MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET", default=True
)
MINIO_PRESIGNED_TTL_MINUTES = env.int("MINIO_PRESIGNED_TTL_MINUTES", default=15)
MINIO_DISABLE_CERT_CHECKS = env.bool("MINIO_DISABLE_CERT_CHECKS", default=False)


# GraphQL

GRAPHENE = {
    "SCHEMA": "caluma.schema.schema",
    "MIDDLEWARE": [],
    "RELAY_CONNECTION_MAX_LIMIT": None,
}

# OpenID connect

OIDC_USERINFO_ENDPOINT = env.str("OIDC_USERINFO_ENDPOINT", default=None)
OIDC_VERIFY_SSL = env.bool("OIDC_VERIFY_SSL", default=True)
OIDC_GROUPS_CLAIM = env.str("OIDC_GROUPS_CLAIM", default="caluma_groups")
OIDC_USERNAME_CLAIM = env.str("OIDC_USERNAME_CLAIM", default="sub")
OIDC_BEARER_TOKEN_REVALIDATION_TIME = env.int(
    "OIDC_BEARER_TOKEN_REVALIDATION_TIME", default=0
)

OIDC_INTROSPECT_ENDPOINT = env.str("OIDC_INTROSPECT_ENDPOINT", default=None)
OIDC_INTROSPECT_CLIENT_ID = env.str("OIDC_INTROSPECT_CLIENT_ID", default=None)
OIDC_INTROSPECT_CLIENT_SECRET = env.str("OIDC_INTROSPECT_CLIENT_SECRET", default=None)

CALUMA_OIDC_USER_FACTORY = env.str(
    "CALUMA_OIDC_USER_FACTORY", default="caluma.caluma_user.models.OIDCUser"
)


# Extensions

VISIBILITY_CLASSES = env.list(
    "VISIBILITY_CLASSES", default=default(["caluma.caluma_core.visibilities.Any"])
)

PERMISSION_CLASSES = env.list(
    "PERMISSION_CLASSES", default=default(["caluma.caluma_core.permissions.AllowAny"])
)

VALIDATION_CLASSES = env.list("VALIDATION_CLASSES", default=[])

DATA_SOURCE_CLASSES = env.list("DATA_SOURCE_CLASSES", default=[])

FORMAT_VALIDATOR_CLASSES = env.list("FORMAT_VALIDATOR_CLASSES", default=[])

EVENT_RECEIVER_MODULES = env.list("EVENT_RECEIVER_MODULES", default=[])

DYNAMIC_GROUPS_CLASSES = env.list("DYNAMIC_GROUPS_CLASSES", default=[])

DYNAMIC_TASKS_CLASSES = env.list("DYNAMIC_TASKS_CLASSES", default=[])

# simple history
SIMPLE_HISTORY_HISTORY_ID_USE_UUID = True

# Historical API
ENABLE_HISTORICAL_API = env.bool("ENABLE_HISTORICAL_API", default=False)

# Configure the fields you intend to use in the "meta" fields. This will
# provide corresponding constants in the ordreBy filter, as well as allow
# you to use those fields in the analytics module.
META_FIELDS = env.list("META_FIELDS", default=[])

# enable caluma healthz endpoint
ENABLE_HEALTHZ_ENDPOINT = env.bool("ENABLE_HEALTHZ_ENDPOINT", default=False)



import itertools

from django.apps import apps
from django.core import serializers
from django.core.management.base import BaseCommand, CommandError
from django.db import DEFAULT_DB_ALIAS
from django.db.models import Q


def get_form_filters(form_slugs):
    return {
        "caluma_form.Form": Q(pk__in=form_slugs),
        "caluma_form.Option": Q(questions__forms__pk__in=form_slugs),
        "caluma_form.Question": Q(forms__pk__in=form_slugs),
        "caluma_form.QuestionOption": Q(question__forms__pk__in=form_slugs),
        "caluma_form.FormQuestion": Q(form__pk__in=form_slugs),
        "caluma_form.Answer": Q(
            question__forms__pk__in=form_slugs, document__isnull=True
        ),
    }


class Command(BaseCommand):
    help = "Dump caluma objects. Currently supported: caluma_form.Form"

    def add_arguments(self, parser):
        parser.add_argument("args", nargs="*", help="Primary keys of given model")
        parser.add_argument(
            "--model", type=str, help="Model to dump, eg. 'caluma_form.Form'"
        )
        parser.add_argument(
            "--format",
            default="json",
            help="Specifies the output serialization format for fixtures.",
        )
        parser.add_argument(
            "--indent",
            type=int,
            default=2,
            help="Specifies the indent level to use when pretty-printing output.",
        )
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help="Nominates a specific database to dump fixtures from. "
            'Defaults to the "default" database.',
        )
        parser.add_argument(
            "-o", "--output", help="Specifies file to which the output is written."
        )

    def handle(self, *pks, **options):
        format = options["format"]
        indent = options["indent"]
        stream = open(options["output"], "w") if options["output"] else None

        if not pks:
            raise CommandError("Must provide primary keys")

        if format not in serializers.get_public_serializer_formats():
            try:
                serializers.get_serializer(format)
            except serializers.SerializerDoesNotExist:
                pass

            raise CommandError("Unknown serialization format: %s" % format)

        if not options.get("model"):
            raise CommandError("Missing `--model` option")

        app_label, model_label = options["model"].split(".")
        querysets = self.get_objects(pks)

        self.stdout.ending = None
        serializers.serialize(
            format,
            itertools.chain(*querysets),
            indent=indent,
            stream=stream or self.stdout,
        )

        if stream:
            stream.close()

    def get_objects(
        self, pks, app_label="caluma_form", model_label="Form", excluded_pks=None
    ):
        """Collect all objects including their "downwards" dependencies recursively."""

        querysets = []
        excluded_pks = excluded_pks or []
        sub_forms = []

        for model_identifier, model_filter in get_form_filters(pks).items():
            app_label, model_label = model_identifier.split(".")
            model = apps.get_model(app_label, model_label)

            qs = model.objects.filter(model_filter)
            querysets.append(qs)

            if model_label == "Question":
                sub_forms += qs.filter(row_form__isnull=False).values_list(
                    "row_form_id", flat=True
                )
                sub_forms += qs.filter(sub_form__isnull=False).values_list(
                    "sub_form_id", flat=True
                )

        if sub_forms:
            querysets += self.get_objects(set(sub_forms), excluded_pks=excluded_pks)

        return querysets



from factory import Faker, SubFactory

from ..caluma_core.factories import DjangoModelFactory
from . import models


class AnalyticsTableFactory(DjangoModelFactory):
    slug = Faker("slug")
    name = Faker("multilang", faker_provider="name")
    starting_object = Faker(
        "word",
        ext_word_list=[i for i, _ in models.AnalyticsTable.STARTING_OBJECT_CHOICES],
    )

    meta = {}

    class Meta:
        model = models.AnalyticsTable


class AnalyticsFieldFactory(DjangoModelFactory):
    alias = Faker("slug")
    meta = {}
    data_source = Faker("slug")
    table = SubFactory(AnalyticsTableFactory)

    class Meta:
        model = models.AnalyticsField



import dateparser
from django.core.management.base import BaseCommand
from django.utils import timezone

from caluma.caluma_logging.models import AccessLog


class Command(BaseCommand):
    """Cleanup access log."""

    help = "Cleanup access log."

    def add_arguments(self, parser):
        parser.add_argument("--force", dest="force", default=False, action="store_true")
        parser.add_argument(
            "-k",
            "--keep",
            dest="keep",
            default="2 weeks",
            help=(
                "Duration for which to keep the access log, older entries will be removed."
            ),
        )

    def handle(self, *args, **options):
        force = options["force"]
        keep = options["keep"]
        lt = dateparser.parse(options["keep"])

        if lt is not None:
            lt = timezone.make_aware(lt)

        entries = AccessLog.objects.filter(timestamp__lt=lt)

        self.stdout.write(
            f"Deleting {entries.count()} access log entries older than {keep}."
        )

        if force:
            entries.delete()



from collections import OrderedDict
from functools import singledispatch

import graphene
from django.core.exceptions import ImproperlyConfigured
from django.http import Http404
from django.shortcuts import get_object_or_404
from graphene.relay.mutation import ClientIDMutation
from graphene.types import Field, InputField
from graphene.types.mutation import MutationOptions
from graphene.types.objecttype import yank_fields_from_attrs
from graphene_django.converter import convert_django_field, convert_field_to_string
from graphene_django.registry import get_global_registry
from graphene_django.rest_framework.mutation import fields_for_serializer
from graphql.language import ast
from localized_fields.fields import LocalizedField
from rest_framework import exceptions

from .relay import extract_global_id


class ASTAnalyzer:
    @staticmethod
    def get_mutation_params(info):
        @singledispatch
        def _parse_ast_value(arg, info):  # pragma: no cover
            raise RuntimeError(f"Unknown arg {arg}")

        @_parse_ast_value.register(ast.ObjectValueNode)
        def _(arg, info):
            return {
                field.name.value: _parse_ast_value(field.value, info)
                for field in arg.fields
            }

        @_parse_ast_value.register(ast.ListValueNode)
        def _(arg, info):
            return [_parse_ast_value(val, info) for val in arg.values]

        @_parse_ast_value.register(ast.StringValueNode)
        def _(arg, info):
            return arg.value

        @_parse_ast_value.register(ast.VariableNode)
        def _(arg, info):
            return info.variable_values[arg.name.value]

        # No need to keep the registered singledispatch implementation
        # in the namespace
        del _

        current_sel = [
            sel
            for sel in info.operation.selection_set.selections
            if sel.name.value == info.field_name
            # if aliases are used, we need to compare them as well
            and (sel.alias is None or sel.alias.value == info.path.key)
        ][0]

        return {
            arg.name.value: _parse_ast_value(arg.value, info)
            for arg in current_sel.arguments
        }


convert_django_field.register(LocalizedField, convert_field_to_string)


class MutationOptions(MutationOptions):
    lookup_field = None
    lookup_input_kwarg = None
    model_class = None
    model_operations = ["create", "update"]
    serializer_class = None
    return_field_name = None
    return_field_type = None


class Mutation(ClientIDMutation):
    """
    Caluma specific Mutation solving following upstream issues.

    1. Expose node instead of attributes directly.

    Dependend issues:
    https://github.com/graphql-python/graphene-django/issues/376
    https://github.com/graphql-python/graphene-django/issues/386
    https://github.com/graphql-python/graphene-django/issues/462

    2. Validation should be GraphQL errors

    https://github.com/graphql-python/graphene-django/issues/380

    Goal would be to get rid of this custom class when referenced issues
    have been resolved successfully.


    The following `Meta` attributes control the basic behavior:
    * `lookup_field`: The model field that should be used to for performing object lookup of
      individual model instances. Defaults to 'pk'.
    * `lookup_input_kwarg`: Input argument that should be used for object lookup.
      Defaults to 'lookup_field'
    * `serializer_class`: The serializer class that should be used for validating, deserializing input
      and performing side effect.
    * `model_class`: The model class to lookup instance of. Defaults to model of serializer.
    * `model_operations`: Define which operations are allowed. Defaults to `['create', 'update'].
    * `fields`: Restrict input fields. Defaults to serializer fields.
    * `exclude`: Exclude input fields. Defaults to serializer fields.
    * `return_field_name`: Name of return graph. Defaults to camel cased model class name.
                           Maybe set to False to not return a field at all.
    * `return_field_type`: Type of return graph. Defaults to object type of given model_class.
    """

    class Meta:
        abstract = True

    # will be set in caluma_core.AppConfig.ready hook, see apps.py
    # to avoid recursive import error
    permission_classes = None

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        lookup_field=None,
        lookup_input_kwarg=None,
        serializer_class=None,
        model_class=None,
        model_operations=None,
        fields=(),
        exclude=(),
        return_field_name=None,
        return_field_type=None,
        **options,
    ):
        model_operations = (
            model_operations if model_operations else ["create", "update"]
        )
        if not serializer_class:
            raise Exception("serializer_class is required for the Mutation")

        if "update" not in model_operations and "create" not in model_operations:
            raise Exception('model_operations must contain "create" and/or "update"')

        serializer = serializer_class()
        if model_class is None:
            serializer_meta = getattr(serializer_class, "Meta", None)
            if serializer_meta:
                model_class = getattr(serializer_meta, "model", None)

        if lookup_field is None and model_class:
            lookup_field = model_class._meta.pk.name
        if lookup_input_kwarg is None:
            lookup_input_kwarg = lookup_field

        input_fields = fields_for_serializer(serializer, fields, exclude, is_input=True)

        if return_field_name is None:
            model_name = model_class.__name__
            return_field_name = model_name[:1].lower() + model_name[1:]

        if not return_field_type:
            registry = get_global_registry()
            return_field_type = registry.get_type_for_model(model_class)

        output_fields = OrderedDict()
        if return_field_name:
            output_fields[return_field_name] = graphene.Field(return_field_type)

        _meta = MutationOptions(cls)
        _meta.lookup_field = lookup_field
        _meta.lookup_input_kwarg = lookup_input_kwarg
        _meta.model_operations = model_operations
        _meta.serializer_class = serializer_class
        _meta.model_class = model_class
        _meta.fields = yank_fields_from_attrs(output_fields, _as=Field)
        _meta.return_field_name = return_field_name
        _meta.return_field_type = return_field_type

        input_fields = yank_fields_from_attrs(input_fields, _as=InputField)
        super().__init_subclass_with_meta__(
            _meta=_meta, input_fields=input_fields, **options
        )

    @classmethod
    def get_serializer_kwargs(cls, root, info, **input):
        model_class = cls._meta.model_class
        return_field_type = cls._meta.return_field_type

        if model_class:
            instance = cls.get_object(
                root,
                info,
                return_field_type.get_queryset(
                    model_class.objects.select_related(), info
                ),
                **input,
            )
            return {
                "instance": instance,
                "data": input,
                "context": {"request": info.context, "info": info, "mutation": cls},
            }

        return {"data": input, "context": {"request": info.context, "info": info}}

    @classmethod
    def get_object(cls, root, info, queryset, **input):
        lookup_field = cls._meta.lookup_field
        lookup_input_kwarg = cls._meta.lookup_input_kwarg

        if "update" in cls._meta.model_operations and lookup_input_kwarg in input:
            instance = get_object_or_404(
                queryset, **{lookup_field: extract_global_id(input[lookup_input_kwarg])}
            )
        elif "create" in cls._meta.model_operations:
            instance = None
        else:
            raise Exception(
                'Invalid update operation. Input parameter "{0}" required.'.format(
                    lookup_field
                )
            )

        return instance

    @classmethod
    def check_permissions(cls, root, info):
        if cls.permission_classes is None:
            raise ImproperlyConfigured(
                "check that app `caluma.caluma_core` is part of your `INSTALLED_APPS` "
                "or custom mutation has `permission_classes` properly assigned."
            )

        for permission_class in cls.permission_classes:
            if not permission_class().has_permission(cls, info):
                raise exceptions.PermissionDenied()

    @classmethod
    def check_object_permissions(cls, root, info, instance):
        for permission_class in cls.permission_classes:
            if not permission_class().has_object_permission(cls, info, instance):
                raise exceptions.PermissionDenied()

    @classmethod
    def mutate_and_get_payload(cls, root, info, **input):
        cls.check_permissions(root, info)
        kwargs = cls.get_serializer_kwargs(root, info, **input)
        instance = kwargs.get("instance")
        if instance is not None:
            cls.check_object_permissions(root, info, kwargs.get("instance"))

        serializer = cls._meta.serializer_class(**kwargs)

        # TODO: use extensions of error to define what went wrong in validation
        # also see https://github.com/graphql-python/graphql-core/pull/204
        # potentially split each validation error into on GraphQL error
        serializer.is_valid(raise_exception=True)
        return cls.perform_mutate(serializer, info)

    @classmethod
    def perform_mutate(cls, serializer, info):

        obj = serializer.save()
        kwargs = {}
        if cls._meta.return_field_name:
            kwargs[cls._meta.return_field_name] = obj
        return cls(**kwargs)

    @classmethod
    def get_params(cls, info):
        return ASTAnalyzer.get_mutation_params(info)


class UserDefinedPrimaryKeyMixin(object):
    """
    Allows a primary key to be overwritten by user.

    TODO: verify whether this makes sense to send upstream.
    """

    class Meta:
        abstract = True

    @classmethod
    def get_object(cls, root, info, queryset, **input):
        lookup_field = cls._meta.lookup_field
        lookup_input_kwarg = cls._meta.lookup_input_kwarg
        model_class = cls._meta.model_class

        filter_kwargs = {lookup_field: input[lookup_input_kwarg]}
        instance = queryset.filter(**filter_kwargs).first()

        if instance is None and model_class.objects.filter(**filter_kwargs).exists():
            # disallow editing of instances which are not visible by current user
            raise Http404(
                "No %s matches the given query." % queryset.model._meta.object_name
            )

        if instance is None and "create" not in cls._meta.model_operations:
            raise exceptions.ValidationError("Create model operation not allowed.")

        if instance is not None and "update" not in cls._meta.model_operations:
            raise exceptions.ValidationError("Update model operation not allowed.")

        return instance



import itertools

from django.db.models.signals import (
    m2m_changed,
    post_delete,
    post_init,
    post_save,
    pre_delete,
    pre_save,
)
from django.dispatch import receiver
from simple_history.signals import pre_create_historical_record

from caluma.caluma_core.events import filter_events
from caluma.utils import disable_raw

from . import models, structure
from .jexl import QuestionJexl


@receiver(pre_create_historical_record, sender=models.HistoricalAnswer)
def set_historical_answer_type(sender, **kwargs):
    """Receiver for historical answer pre_save signal setting question type."""
    historical_answer = kwargs["history_instance"]

    try:
        # can take the current question, assuming the question type is not changed in a datarace
        historical_answer.history_question_type = historical_answer.question.type
    except models.Question.DoesNotExist:
        # This happens when deleting a question with a default_answer. As the Answer
        # has to exist for this to happen, we can safely assume to already have a
        # historical record of it. So we're taking the question type from the previous
        # revision.
        previous = historical_answer.get_previous_by_history_date()
        historical_answer.history_question_type = previous.history_question_type


@receiver(post_delete, sender=models.Question)
@filter_events(lambda instance: instance.default_answer is not None)
def cleanup_default_answer(sender, instance, **kwargs):
    """Ensure default_answers are cleanedup."""
    instance.default_answer.delete()


@receiver(post_init, sender=models.Document)
@filter_events(lambda instance: instance.family_id is None)
def set_document_family(sender, instance, **kwargs):
    """
    Ensure document has the family pointer set.

    This sets the document's family if not overruled or set already.
    The family will be in the corresponding mutations when a tree
    structure is created or restructured.
    """

    # can't use instance.set_family() here as the instance isn't
    # in DB yet, causing integrity errors
    instance.family = instance


def _update_calc_dependents(slug, old_expr, new_expr):
    jexl = QuestionJexl()
    old_q = set(
        list(jexl.extract_referenced_questions(old_expr))
        + list(jexl.extract_referenced_mapby_questions(old_expr))
    )
    new_q = set(
        list(jexl.extract_referenced_questions(new_expr))
        + list(jexl.extract_referenced_mapby_questions(new_expr))
    )

    to_add = new_q - old_q
    to_remove = old_q - new_q

    questions = models.Question.objects.filter(pk__in=list(to_add | to_remove))

    for question in questions:
        deps = set(question.calc_dependents)
        if question.slug in to_add:
            deps.add(slug)
        else:
            deps.remove(slug)
        question.calc_dependents = list(deps)
        question.save()


def _update_or_create_calc_answer(question, document):
    root_doc = document.family

    struc = structure.FieldSet(root_doc, root_doc.form)
    field = struc.get_field(question.slug)

    # skip if question doesn't exist in this document structure
    if field is None:
        return

    jexl = QuestionJexl(
        {"form": field.form, "document": field.document, "structure": field.parent()}
    )

    # Ignore errors because we evaluate greedily as soon as possible. At
    # this moment we might be missing some answers or the expression might
    # be invalid, in which case we return None
    value = jexl.evaluate(field.question.calc_expression, raise_on_error=False)

    models.Answer.objects.update_or_create(
        question=question, document=field.document, defaults={"value": value}
    )


# Update calc dependents on pre_save
#
# Every question that is referenced in a `calcExpression` will memoize the
# referencing calculated question. This list of calculation dependents is
# mutated whenever a calculated question is created, updated or deleted.


@receiver(pre_save, sender=models.Question)
@disable_raw
@filter_events(lambda instance: instance.type == models.Question.TYPE_CALCULATED_FLOAT)
def save_calc_dependents(sender, instance, **kwargs):

    original = models.Question.objects.filter(pk=instance.pk).first()
    if not original:
        _update_calc_dependents(
            instance.slug, old_expr="false", new_expr=instance.calc_expression
        )

    elif original.calc_expression != instance.calc_expression:
        _update_calc_dependents(
            instance.slug,
            old_expr=original.calc_expression,
            new_expr=instance.calc_expression,
        )


@receiver(pre_delete, sender=models.Question)
@filter_events(lambda instance: instance.type == models.Question.TYPE_CALCULATED_FLOAT)
def remove_calc_dependents(sender, instance, **kwargs):
    _update_calc_dependents(
        instance.slug, old_expr=instance.calc_expression, new_expr="false"
    )


# Update calculated answer on post_save
#
# Try to update the calculated answer value whenever a mutation on a possibly
# related model is performed.


@receiver(post_save, sender=models.Question)
@disable_raw
@filter_events(lambda instance: instance.type == models.Question.TYPE_CALCULATED_FLOAT)
def update_calc_from_question(sender, instance, created, update_fields, **kwargs):

    # TODO optimize: only update answers if calc_expression is updated
    # needs to happen during save() or __init__()

    for document in models.Document.objects.filter(form__questions=instance):
        _update_or_create_calc_answer(instance, document)


@receiver(post_save, sender=models.FormQuestion)
@disable_raw
@filter_events(
    lambda instance: instance.question.type == models.Question.TYPE_CALCULATED_FLOAT
)
def update_calc_from_form_question(sender, instance, created, **kwargs):
    for document in instance.form.documents.all():
        _update_or_create_calc_answer(instance.question, document)


@receiver(post_save, sender=models.Answer)
@disable_raw
@filter_events(lambda instance: instance.document and instance.question.calc_dependents)
def update_calc_from_answer(sender, instance, **kwargs):
    # If there is no document on the answer it means that it's a default
    # answer. They shouldn't trigger a recalculation of a calculated field
    # even when they are technically listed as a dependency.
    # Also skip non-referenced answers.

    for question in models.Question.objects.filter(
        pk__in=instance.question.calc_dependents
    ):
        _update_or_create_calc_answer(question, instance.document)


@receiver(post_save, sender=models.Document)
@disable_raw
@filter_events(lambda created: created)
def update_calc_from_document(sender, instance, created, **kwargs):

    for question in instance.form.questions.filter(
        type=models.Question.TYPE_CALCULATED_FLOAT
    ):
        _update_or_create_calc_answer(question, instance)


@receiver(m2m_changed, sender=models.AnswerDocument)
def update_calc_from_answerdocument(sender, instance, **kwargs):
    dependents = instance.document.form.questions.exclude(
        calc_dependents=[]
    ).values_list("calc_dependents", flat=True)

    dependent_questions = list(itertools.chain(*dependents))

    for question in models.Question.objects.filter(pk__in=dependent_questions):
        _update_or_create_calc_answer(question, instance.document)



# Generated by Django 2.2.24 on 2022-01-17 12:50

import django.db.models.deletion
from django.db import migrations, models

import caluma.caluma_core.models


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_workflow", "0027_add_modified_by_user_group"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicaltaskflow",
            name="redoable",
            field=models.TextField(
                blank=True,
                help_text="jexl returning what tasks can be redone from this taskflow.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="taskflow",
            name="redoable",
            field=models.TextField(
                blank=True,
                help_text="jexl returning what tasks can be redone from this taskflow.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalworkitem",
            name="status",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("ready", "Work item is ready to be processed."),
                    ("completed", "Work item is done."),
                    ("canceled", "Work item is canceled."),
                    ("skipped", "Work item is skipped."),
                    ("suspended", "Work item is suspended."),
                    ("redo", "Work item has been marked for redo."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="taskflow",
            name="flow",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="task_flows",
                to="caluma_workflow.Flow",
            ),
        ),
        migrations.AlterField(
            model_name="workitem",
            name="status",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("ready", "Work item is ready to be processed."),
                    ("completed", "Work item is done."),
                    ("canceled", "Work item is canceled."),
                    ("skipped", "Work item is skipped."),
                    ("suspended", "Work item is suspended."),
                    ("redo", "Work item has been marked for redo."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-06 08:54
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_form", "0003_auto_20190130_0920"),
        ("caluma_workflow", "0006_auto_20190121_1440"),
    ]

    operations = [
        migrations.RemoveField(model_name="workflow", name="form"),
        migrations.AddField(
            model_name="workflow",
            name="allow_all_forms",
            field=models.BooleanField(
                default=False, help_text="Allow workflow to be started with any form"
            ),
        ),
        migrations.AddField(
            model_name="workflow",
            name="allow_forms",
            field=models.ManyToManyField(
                blank=True,
                help_text="List of forms which are allowed to start workflow with",
                related_name="workflows",
                to="caluma_form.Form",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['caluma',
 'caluma.caluma_analytics',
 'caluma.caluma_analytics.management.commands',
 'caluma.caluma_analytics.migrations',
 'caluma.caluma_core',
 'caluma.caluma_core.management',
 'caluma.caluma_core.management.commands',
 'caluma.caluma_data_source',
 'caluma.caluma_form',
 'caluma.caluma_form.management',
 'caluma.caluma_form.management.commands',
 'caluma.caluma_form.migrations',
 'caluma.caluma_logging',
 'caluma.caluma_logging.management',
 'caluma.caluma_logging.management.commands',
 'caluma.caluma_logging.migrations',
 'caluma.caluma_user',
 'caluma.caluma_workflow',
 'caluma.caluma_workflow.migrations',
 'caluma.extensions',
 'caluma.settings']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2.12,<4.0.0',
 'dateparser>=1.1.0,<2.0.0',
 'django-cors-headers>=3.11.0,<4.0.0',
 'django-environ>=0.8.1,<0.9.0',
 'django-extensions>=3.1.5,<4.0.0',
 'django-filter>=21.1,<22.0',
 'django-localized-fields>=6.6,<7.0',
 'django-postgres-extra>=2.0.4,<3.0.0',
 'django-simple-history>=3.0.0,<4.0.0',
 'django-watchman>=1.2.0,<2.0.0',
 'djangorestframework>=3.13.1,<4.0.0',
 'graphene-django>=3.0.0b7,<4.0.0',
 'graphql-relay>=3.1.5,<4.0.0',
 'idna>=3.3,<4.0',
 'minio>=7.1.4,<8.0.0',
 'psycopg2-binary>=2.9.3,<3.0.0',
 'pyjexl>=0.3.0,<0.4.0',
 'python-memcached>=1.59,<2.0',
 'requests>=2.27.1,<3.0.0',
 'uWSGI>=2.0.20,<3.0.0',
 'urllib3>=1.26.8,<2.0.0']

setup_kwargs = {
    'name': 'caluma',
    'version': '8.0.0b0',
    'description': 'Caluma Service providing GraphQL API',
    'long_description': "# ![Caluma Service](https://user-images.githubusercontent.com/6150577/60805422-51b1bf80-a180-11e9-9ae5-c794249c7a98.png)\n\n[![Build Status](https://github.com/projectcaluma/caluma/workflows/Tests/badge.svg)](https://github.com/projectcaluma/caluma/actions?query=workflow%3ATests)\n[![Coverage](https://img.shields.io/badge/coverage-100%25-brightgreen.svg)](https://github.com/projectcaluma/caluma/blob/main/setup.cfg#L57)\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/python/black)\n[![PyPI](https://img.shields.io/pypi/v/caluma)](https://pypi.org/project/caluma/)\n[![License: GPL-3.0-or-later](https://img.shields.io/github/license/projectcaluma/caluma)](https://spdx.org/licenses/GPL-3.0-or-later.html)\n\nCaluma is a collaborative form editing and workflow service.\n\n- Website: [caluma.io](https://caluma.io)\n- Documentation: [caluma.gitbook.io](https://caluma.gitbook.io)\n\n## Getting started\n\n**Requirements**\n\n- docker\n- docker-compose\n\nAfter installing and configuring those, download [docker-compose.yml](https://github.com/projectcaluma/caluma/blob/main/docker-compose.yml) and run the following command:\n\n```bash\ndocker-compose up -d\n```\nSchema introspection and documentation is available at [http://localhost:8000/graphql](localhost:8000/graphql) and can be accessed using a GraphQL client such as [Altair](https://altair.sirmuel.design/). The API allows to query and mutate form and workflow entities which are described below.\n\nYou can read more about running and configuring Caluma in the [documentation](https://caluma.gitbook.io).\n\n## License\n\nCode released under the [GPL-3.0-or-later license](LICENSE).\n\nFor further information on our license choice, you can read up on the [corresponding GitHub issue](https://github.com/projectcaluma/caluma/issues/751#issuecomment-547974930).\n\n---\n\n- Contributing guide: [CONTRIBUTING.md](CONTRIBUTING.md)\n- Maintainer's Handbook: [MAINTAINING.md](MAINTAINING.md)\n",
    'author': 'Caluma',
    'author_email': 'info@caluma.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://caluma.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from graphene import List, String
from graphene_django import forms
from graphene_django.forms.converter import convert_form_field


class GlobalIDFormField(forms.GlobalIDFormField):
    """
    Global id form field which allows plain primary keys.

    Disable cleaning as input may be an actual relay global id
    or plain primary key.
    """

    def clean(self, value):
        return value


class GlobalIDMultipleChoiceField(forms.GlobalIDMultipleChoiceField):
    """
    Global id multiple choice field which allows plain primary keys.

    Disable cleaning as input may be an actual relay global id
    or plain primary key.
    """

    def valid_value(self, value):
        return True


class SlugMultipleChoiceField(GlobalIDMultipleChoiceField):
    """
    Slug multiple choice field which allows listing multiple slugs.

    Disable cleaning as we don't know which slugs may be valid down
    the line
    """

    def valid_value(self, value):  # pragma: no cover
        return True


@convert_form_field.register(SlugMultipleChoiceField)
def slug_to_list(field):
    return List(String, required=field.required)



# Generated by Django 2.2.3 on 2019-07-18 12:35

from django.db import migrations

import caluma.caluma_core.models


class Migration(migrations.Migration):

    dependencies = [
        (
            "caluma_workflow",
            "0012_historicalcase_historicalflow_historicaltask_historicaltaskflow_historicalworkflow_historicalworkite",
        )
    ]

    operations = [
        migrations.AlterField(
            model_name="case",
            name="status",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("running", "Case is running and work items need to be completed."),
                    ("completed", "Case is done."),
                    ("canceled", "Case is cancelled."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalcase",
            name="status",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("running", "Case is running and work items need to be completed."),
                    ("completed", "Case is done."),
                    ("canceled", "Case is cancelled."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicaltask",
            name="type",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("simple", "Task which can simply be marked as completed."),
                    (
                        "complete_workflow_form",
                        "Task to complete a defined workflow form.",
                    ),
                    ("complete_task_form", "Task to complete a defined task form."),
                ],
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalworkitem",
            name="status",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("ready", "Task is ready to be processed."),
                    ("completed", "Task is done."),
                    ("canceled", "Task is cancelled."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="type",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("simple", "Task which can simply be marked as completed."),
                    (
                        "complete_workflow_form",
                        "Task to complete a defined workflow form.",
                    ),
                    ("complete_task_form", "Task to complete a defined task form."),
                ],
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="workitem",
            name="status",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("ready", "Task is ready to be processed."),
                    ("completed", "Task is done."),
                    ("canceled", "Task is cancelled."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
    ]



# Generated by Django 2.2.8 on 2020-02-10 09:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0027_auto_20200113_0727")]

    operations = [
        migrations.AlterField(
            model_name="answer",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="answer",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="answer",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="answerdocument",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="answerdocument",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="answerdocument",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="document",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="document",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="document",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="dynamicoption",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="dynamicoption",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="dynamicoption",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="file",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="file",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="file",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="form",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="form",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="form",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="formquestion",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="formquestion",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="formquestion",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalanswer",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalanswer",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalanswer",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalanswerdocument",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalanswerdocument",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalanswerdocument",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicaldocument",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicaldocument",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicaldocument",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicaldynamicoption",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicaldynamicoption",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicaldynamicoption",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalfile",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalfile",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalfile",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalform",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalform",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalform",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalformquestion",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalformquestion",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalformquestion",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicaloption",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicaloption",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicaloption",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalquestion",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalquestion",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalquestion",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalquestionoption",
            name="created_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalquestionoption",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalquestionoption",
            name="modified_at",
            field=models.DateTimeField(blank=True, db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name="option",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="option",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="option",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="question",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="question",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="question",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="questionoption",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="questionoption",
            name="created_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="questionoption",
            name="modified_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
    ]



from __future__ import annotations  # self-referencing annotations

import re
from copy import deepcopy
from datetime import datetime
from functools import cached_property, partial
from typing import List, Optional

from django.conf import settings
from django.db import connection
from django.utils import timezone
from psycopg2.extras import DictCursor

from caluma.caluma_form import models as form_models
from caluma.caluma_workflow import models as workflow_models

from . import sql


class BaseField:
    """Base class for all (available) fields."""

    def __init__(
        self,
        *,
        parent: Optional[BaseField],
        identifier: str,
        visibility_source,
        label=None,
        **_,
    ):
        """Initialize new field instance.

        Parameters are as follows:
         - `parent`: an optional reference to the superordinate field
         - `identifier`: the DB field that's being extracted
         - `visibility_source`: Helper to fetch visibility-aware querysets
         - `label`: Optional label
        """
        self.parent = parent
        self.identifier = identifier
        self.visibility_source = visibility_source
        self.label = label if label else identifier
        self.filters = None

        # required - set to True if the result of this field is
        # needed in the analysis.
        # Optimisation - SQL generation may be skipped in some
        # cases if field not required
        self.required = False

    def path_args(self):
        """Extract arguments from the field's path."""
        match = re.match(r"^\w+?\[(.*)\]$", self.identifier)
        if match:
            return match.group(1).split(",")
        return []  # pragma: no cover

    @property
    def available_children(self):  # pragma: no cover
        """Return a dict of all available children.

        Used in the construction/design-time stage,
        as well as when building an SQL query from a given
        set of paths.
        """
        return {}

    def full_label(self):
        """Return a full hierarchical path to the given field.

        Includes the full label of the parent field(s) as well
        """
        if self.parent:
            return f"{self.parent.full_label()} > {self.label}"
        return self.label

    def source_path(self) -> List[str]:
        """Return the full source path of this field as a list."""
        return (
            self.parent.source_path() + [self.identifier]
            if self.parent
            else [self.identifier]
        )

    def parse_value(self, value):
        """Parse value from DB into python value.

        Only needs to be overloaded when required.
        Used in GraphQL context only
        """
        return value

    def query_field(self):
        """Return the corresponding SQL rendering field."""
        raise NotImplementedError(
            f"Method 'query_field' is missing in {type(self)}"
        )  # pragma: no cover

    def is_value(self):
        """Return True if this field represents a usable value."""
        raise NotImplementedError(
            f"Method 'is_value' is missing in {type(self)}"
        )  # pragma: no cover

    def is_leaf(self):
        """Return True if there are no more fields below this one."""
        raise NotImplementedError(
            f"Method 'is_leaf' is missing in {type(self)}"
        )  # pragma: no cover


class MetaField(BaseField):
    """MetaField implements access to JSON fields and their contents.

    For this to work correctly, the setting META_FIELDS must be
    configured to list the fields that can be extracted.
    If a key is set in a `meta` value but not listed in the
    settings, it will not be usable in Analytics.
    """

    def __init__(self, parent, identifier, label, *, visibility_source, meta_name=None):
        super().__init__(
            parent=parent,
            identifier=identifier,
            label=label,
            visibility_source=visibility_source,
        )
        self.meta_name = meta_name

    def is_leaf(self):
        # if we have a meta field name, we are a leaf:
        # We cannot descend further down, there are no children
        return bool(self.meta_name)

    def is_value(self):
        # "plain" meta cannot be used as a value
        return not self.is_leaf()

    @cached_property
    def available_children(self):
        if self.meta_name:
            return {}
        return {
            name: MetaField(
                parent=self,
                identifier=self.identifier,
                label=f"Key '{name}'",
                meta_name=name,
                visibility_source=self.visibility_source,
            )
            for name in settings.META_FIELDS
        }

    def source_path(self):
        source_path = super().source_path()
        if self.meta_name:
            # replace last element with our "meta" name
            # so path is correct, and identifier still can point
            # to the actual DB field.
            source_path[-1] = self.meta_name
        return source_path

    def query_field(self):
        if self.meta_name:
            return sql.JSONExtractorField(
                self.identifier,
                parent=self.parent.query_field() if self.parent else None,
                json_key=self.meta_name,
            )
        return sql.NOOPField(
            self.identifier,
            parent=self.parent.query_field() if self.parent else None,
        )


class AttributeField(BaseField):
    """Extractor for regular table attributes.

    When `is_date` is set, it will also contain a number of sub-fields
    that represent extracts of some date parts (such as year, quarter,
    month, and weekday).
    """

    def __init__(self, parent, identifier, *, visibility_source, is_date=False):
        super().__init__(
            parent=parent, identifier=identifier, visibility_source=visibility_source
        )
        self.is_date = is_date

    def is_leaf(self):
        return not self.is_date

    def is_value(self):
        return True

    def query_field(self):
        if self.is_date and not self.required:
            return sql.NOOPField(
                self.identifier,
                parent=self.parent.query_field() if self.parent else None,
            )
        return sql.AttrField(
            self.identifier,
            parent=self.parent.query_field() if self.parent else None,
        )

    @cached_property
    def available_children(self):
        if self.is_date:
            # for FormAnswerField, the identifier is the question slug,
            # but the field is 'date'
            date_field = getattr(self, "attr_name", self.identifier)

            extractor_field = partial(
                DateExtractorField,
                parent=self,
                date_field=date_field,
                visibility_source=self.visibility_source,
            )
            return {
                "year": extractor_field(identifier="year"),
                "month": extractor_field(identifier="month"),
                "weekday": extractor_field(identifier="weekday"),
                "quarter": extractor_field(identifier="quarter"),
            }
        return {}

    def parse_value(self, value):
        if not self.is_date:
            return value

        if isinstance(value, datetime):
            # psycopg2 returns "generic" tz info dates that sometimes
            # are weird when comparing to the ones Django expects
            # (Note they're still correct, just not labeled in an
            # useful way)
            current_tz = timezone.get_current_timezone()
            return current_tz.normalize(value)
        return value


class WorkItemField(BaseField):
    """Field that provides access to a work item's data."""

    ORDER_TYPES = {
        "first": ["created_at"],
        "last": ["created_at DESC"],
        "firstclosed": ["closed_at"],
        "lastclosed": ["closed_at DESC"],
    }
    ORDER_LABELS = {
        "first": "first created",
        "last": "last created",
        "firstclosed": "first closed",
        "lastclosed": "first closed",
    }

    def __init__(
        self,
        parent,
        identifier,
        visibility_source,
        task_slug=None,
        ordering_selector=None,
    ):
        super().__init__(
            parent=parent, identifier=identifier, visibility_source=visibility_source
        )

        self.task_slug = task_slug if task_slug else self.path_args()[0]
        self.ordering_selector = (
            ordering_selector if ordering_selector else self.path_args()[1]
        )
        assert self.ordering_selector in self.ORDER_TYPES

        # TODO translate?
        self.label = f"Workitem: {self.ORDER_LABELS[self.ordering_selector]} with task {self.task_slug}"

    def _order_by(self):
        return self.ORDER_TYPES[self.ordering_selector]

    def is_leaf(self):
        return False

    def is_value(self):
        return False

    def query_field(self):
        return sql.JoinField(
            self.identifier,
            table=self.visibility_source.workitems(),
            outer_ref=("id", "case_id"),
            filters=[f"task_id = '{self.task_slug}'"],
            order_by=self._order_by(),
            parent=self.parent.query_field() if self.parent else None,
        )

    @cached_property
    def available_children(self):
        direct_fields = {
            "meta": MetaField(
                parent=self,
                identifier="meta",
                label="Meta",
                visibility_source=self.visibility_source,
            ),
            "closed_at": AttributeField(
                parent=self,
                identifier="closed_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "deadline": AttributeField(
                parent=self,
                identifier="closed_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "created_at": AttributeField(
                parent=self,
                identifier="created_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "status": AttributeField(
                parent=self,
                identifier="status",
                visibility_source=self.visibility_source,
            ),
            "name": AttributeField(
                parent=self, identifier="name", visibility_source=self.visibility_source
            ),
        }

        wi_forms = form_models.Form.objects.filter(
            documents__work_item__isnull=False
        ).distinct()

        form_fields = {
            f"document[{form.slug}]": FormDocumentField(
                identifier=f"document[{form.slug}]",
                parent=self,
                visibility_source=self.visibility_source,
            )
            for form in wi_forms
        }

        return {**direct_fields, **form_fields}


class FormDocumentField(BaseField):
    """Field that provides access to a form's data."""

    def __init__(
        self, parent, identifier, visibility_source, subform_level=0, form_slug=None
    ):
        super().__init__(
            parent=parent, identifier=identifier, visibility_source=visibility_source
        )

        self.form_slug = form_slug if form_slug else self.path_args()[0]

        # subform level is used so we can have the "path" as it is presented
        # to the user (our `location`) separate from where the corresponding
        # `Answer` is found (remember, answers of questions in subforms are
        # NOT put into sub-documents according to hierarchy, but stored in a
        # single document!)
        self.subform_level = subform_level

    def query_field(self):
        # Only for top-level form do we need a Join field.
        # The "sub-forms"' answers are all on the same document,
        # thus for those we return a NOOP field.
        if self.subform_level == 0:
            return sql.JoinField(
                self.identifier,
                table=self.visibility_source.documents(),
                outer_ref=("document_id", "id"),
                filters=[f"form_id = '{self.form_slug}'"],
                # TODO HMM this is wrong, join field should always have a parent
                parent=self.parent.query_field() if self.parent else None,
            )
        # else - we just return the field from our "virtual" parent
        return self.parent.query_field()

    def is_leaf(self):
        return False

    def is_value(self):
        return False

    @cached_property
    def available_children(self):
        form = form_models.Form.objects.get(pk=self.form_slug)
        questions = form.questions.all().exclude(type=form_models.Question.TYPE_TABLE)
        children = {}

        for question in questions:
            qf = self._question_field(question)
            if qf:
                children[question.pk] = qf
        return children

    def _question_field(self, question):
        # Note: we explicitly don't support table questions at the moment.
        # They would require another sub-selector to get the first, or last
        # item in the table, which in most cases in real-world use isn't very
        # useful.
        if question.type in (
            question.TYPE_INTEGER,
            question.TYPE_FLOAT,
            question.TYPE_TEXT,
            question.TYPE_TEXTAREA,
            question.TYPE_CHOICE,
            question.TYPE_DYNAMIC_CHOICE,
        ):
            return FormAnswerField(
                parent=self,
                identifier=question.slug,
                attr_name="value",
                subform_level=self.subform_level,
                visibility_source=self.visibility_source,
            )
        elif question.type == question.TYPE_DATE:
            return FormAnswerField(
                parent=self,
                identifier=question.slug,
                attr_name="date",
                is_date=True,
                subform_level=self.subform_level,
                visibility_source=self.visibility_source,
            )

        elif question.type == question.TYPE_FORM:
            return FormDocumentField(
                parent=self,
                identifier=question.slug,
                form_slug=question.sub_form_id,
                subform_level=self.subform_level + 1,
                visibility_source=self.visibility_source,
            )


class FormAnswerField(AttributeField):
    """Represents access to an answer within a form."""

    def __init__(self, parent, identifier, attr_name, subform_level, **kwargs):
        super().__init__(parent=parent, identifier=identifier, **kwargs)
        self.subform_level = subform_level
        self.attr_name = attr_name

    def query_field(self):
        # Only for top-level form do we need a Join field.
        # The "sub-forms"' answers are all on the same document,
        # thus for those we return a NOOP field.
        answer_join_field = sql.JoinField(
            identifier=self.identifier,
            table=self.visibility_source.answers(),
            # TODO: turn this into SQL parameter
            filters=[f""""question_id" = '{self.identifier}' """],
            outer_ref=("id", "document_id"),
            parent=self.parent.query_field() if self.parent else None,
        )
        value_field = sql.AttrField(
            identifier=self.attr_name,
            parent=answer_join_field,
            filter_values=self.filters,
            answer_value_mode=(self.attr_name == "value"),
        )
        return value_field


class DateExtractorField(AttributeField):
    """Specialisation of attribute field: Extract a date part."""

    def __init__(self, parent, identifier, visibility_source, date_field, **kwargs):
        super().__init__(
            parent,
            identifier,
            visibility_source=visibility_source,
            is_date=False,
            **kwargs,
        )
        self.date_field = date_field

    def query_field(self):
        return sql.DateExprField(
            self.date_field,
            parent=self.parent.query_field() if self.parent else None,
            extract_part=self.identifier,
            filter_values=self.filters,
        )

    def parse_value(self, value):
        """Convert extracted date part to integer.

        This is necessary because not all supported postgres versions return the
        same datatype which leads to inconsistency. E.g versions < 12 returns a
        float while later versions return a decimal.
        """
        return int(value)


class BaseStartingObject:
    """Base class for starting objects.

    Provides interface for listing the fields within the given starting
    object, and also supports Enum generation on the GraphQL side.
    """

    def __init__(self, info, disable_visibilities):
        """Initialize starting object.

        The `info` parameter is the Graphql info/context object.
        """
        self.info = info

        # deferred import to avoid circular deps
        from . import visibility

        self.visibility_source = visibility.CalumaVisibilitySource(
            info, is_disabled=disable_visibilities
        )

    @classmethod
    def get_object(cls, identifier, info, disable_visibilities=False):
        """Return starting object identified by given name.

        Call `BaseStartingObject.get_object('cases', info)`
        to get a CaseStartingObject, for example.
        """

        subclasses_by_name = {sub.identifier: sub for sub in cls.__subclasses__()}
        return subclasses_by_name[identifier](
            info, disable_visibilities=disable_visibilities
        )

    @classmethod
    def as_choices(cls):
        """Return a list of (slug, label) tuples of all available starting objects.

        This can then be used in a CharField's choices.
        """

        objects = sorted(
            cls.__subclasses__(),
            key=lambda obj: obj.label,
        )
        return [(obj.identifier, obj.label) for obj in objects]

    def get_fields(self, prefix=None, depth=1):
        """Return fields for the GraphQL interface.

        When not parametrized, return the root fields of the starting object.

        When a given prefix and depth are given, return the fields from the
        specified sub-path, and go down the requested depth.
        """
        if prefix:
            # support string prefixes as well
            prefix = prefix if isinstance(prefix, list) else prefix.split(".")

        def _children_at_depth(field, _depth):
            if _depth <= 0:
                return
            for path, child in field.available_children.items():
                yield ([path], child)
                for subpath, grandchild in _children_at_depth(child, _depth - 1):
                    yield ([path] + subpath, grandchild)

        if prefix:
            # get only fields with given prefix, and start looking
            # down `depth` levels from there
            fields = self.get_field(prefix).available_children
        else:
            prefix = []
            fields = self.root_fields()

        output = {}
        for path, field in fields.items():

            output[".".join(prefix + [path])] = field
            for subpath, child in _children_at_depth(field, depth - 1):
                output[".".join(prefix + [path] + subpath)] = child
        return output

    def root_fields(self):  # pragma: no cover
        """Return the root fields for this starting object.

        Must be overridden in a subclass actually implementing
        a starting object.
        """
        raise NotImplementedError("Must be implemented in BaseStartingObject subclass")

    def get_field(self, path):
        """Return a single field with the given path."""
        children = self.root_fields()
        for elem in path:
            field = children[elem]
            children = field.available_children
        return field


class CaseStartingObject(BaseStartingObject):
    identifier = "cases"
    label = "Cases"

    def get_query(self):
        return self.visibility_source.cases()

    def root_fields(self):
        direct_fields = {
            "meta": MetaField(
                parent=None,
                identifier="meta",
                label="Meta",
                visibility_source=self.visibility_source,
            ),
            "closed_at": AttributeField(
                parent=None,
                identifier="closed_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "created_at": AttributeField(
                parent=None,
                identifier="created_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "status": AttributeField(
                parent=None,
                identifier="status",
                visibility_source=self.visibility_source,
            ),
            "id": AttributeField(
                parent=None,
                identifier="id",
                visibility_source=self.visibility_source,
            ),
        }

        case_forms = form_models.Form.objects.filter(
            documents__case__isnull=False
        ).distinct()

        form_fields = {
            f"document[{form.slug}]": FormDocumentField(
                identifier=f"document[{form.slug}]",
                parent=None,
                visibility_source=self.visibility_source,
            )
            for form in case_forms
        }

        tasks = workflow_models.Task.objects.all()

        workitem_fields = {
            f"workitem[{task.slug},{selector}]": WorkItemField(
                identifier=f"workitem[{task.slug},{selector}]",
                parent=None,
                visibility_source=self.visibility_source,
            )
            for selector in ["first", "last", "firstclosed", "lastclosed"]
            for task in tasks
        }

        return {**direct_fields, **form_fields, **workitem_fields}


class SimpleTable:
    class _AnonymousInfo:
        """
        Pseudo GraphQL info object.

        Used for commandline access, so the commandline version
        can also invoke the visibilities.
        """

        context = {}

    def __init__(self, table, info=_AnonymousInfo):
        self.info = info

        self.table = table
        self.last_query = None
        self.last_query_params = None

        self.starting_object = BaseStartingObject.get_object(
            self.table.starting_object,
            self.info,
            disable_visibilities=self.table.disable_visibilities,
        )

        self.base_query = self.starting_object.get_query()

    @cached_property
    def _fields(self):
        # sort fields to join by length. This way, parent fields
        # will be already joined when their children get added, so
        # they will have the right alias.
        table_fields = sorted(self.table.fields.all(), key=lambda f: len(f.data_source))

        fields = {}
        for field_spec in table_fields:
            fields[field_spec.alias] = self.starting_object.get_field(
                field_spec.data_source.split(".")
            )
            fields[field_spec.alias].alias = field_spec.alias

            # Filters must always be on the outermost (base) query,
            # as we're using LEFT JOINs to get at related data.
            self.base_query.add_field_filter(
                self._sql_alias(field_spec.alias), field_spec.filters
            )
        return fields

    def get_sql_and_params(self):
        """Return a list of records as specified in the given table config."""

        fields = self._fields

        step_queries = {}
        base_query = deepcopy(self.base_query)

        for alias, field in fields.items():
            query = base_query

            field.required = True
            sql_field = field.query_field()
            sql_field.alias = self._sql_alias(alias)

            cache_path = ""
            for step in sql_field.path_from_root():
                cache_path = f"{cache_path}.{step.identifier}"
                if cache_path in step_queries:
                    query = step_queries[cache_path]
                else:
                    query = step.annotate(query)
                    step_queries[cache_path] = query

        sql_query, params, _ = sql.QueryRender(base_query).as_sql(alias=None)

        self.last_query = sql_query
        self.last_query_params = params
        return sql_query, params

    def get_records(self):

        sql_query, params = self.get_sql_and_params()

        with connection.connection.cursor(cursor_factory=DictCursor) as cursor:
            cursor.execute(sql_query, params)
            data = cursor.fetchall()

            return [
                {
                    field.alias: field.parse_value(
                        row[f"analytics_result_{field.alias}"]
                    )
                    for field in self._fields.values()
                }
                for row in data
            ]

    def _sql_alias(self, user_alias):
        return f"analytics_result_{user_alias}"
